import os
import pytest
import sys
from astro_ghost.PS1QueryFunctions import *
from astro_ghost.TNSQueryFunctions import getTNSSpectra
from astro_ghost.NEDQueryFunctions import *
from astro_ghost.ghostHelperFunctions import *
from astro_ghost.stellarLocus import *
from astropy.coordinates import SkyCoord
from astropy import units as u
import pandas as pd
from datetime import datetime
import astro_ghost

#we want to include print statements so we know what the algorithm is doing
verbose = 1


#need to generate tests for photoz_helper and gradientAscent!!

t = np.linspace(0, 70)
y = 7.*2.71828**(-(t)/50)/(1+2.71828**(-(t)/11))
plt.plot(t, y)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config


# -- Project information -----------------------------------------------------

project = 'astro_ghost'
copyright = '2021, Alexander Gagliano'
author = 'Alexander Gagliano'

# The full version, including alpha/beta/rc tags
from astro_ghost import __version__
release = __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.mathjax',
    'sphinx_copybutton',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'sphinx_tabs.tabs',
    'sphinx-prompt',
    'sphinx_toolbox.shields',
    'sphinx_toolbox.github',
]

github_username = 'alexandergagliano'
github_repository = 'uiucsn/astro_ghost'

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



from bs4 import BeautifulSoup
import requests
import urllib.request
from astro_ghost.PS1QueryFunctions import find_all
from astro_ghost.ghostHelperFunctions import remove_prefix

def getTNSSpectra(transients, path, verbose=False):
    """Scrapes the Transient Name Server for all public spectra of a transient.

    :param transients: Dataframe of transient information (must contain the column 'Name').
    :type transients: Pandas DataFrame
    :param path: Filepath where spectra will be saved.
    :type path: str
    :param verbose: If True, print progress.
    :type verbose: bool, optional
    """

    names = [remove_prefix(x, "SN") for x in transients['Name']]

    # loop through transients
    for name in names:
        # set up URL for TNS object
        URL = 'https://wis-tns.weizmann.ac.il/object/' + name
        page = requests.get(URL)
        soup = BeautifulSoup(page.content, 'html.parser')
        files = soup.findAll("td", {"class": "cell-asciifile"})

        if (len(files) > 0):
            if verbose:
                print("%s found in TNS. Downloading spectra:"%name)
            for file in files:
                link = file.a['href']
                fn = link.split("/")[-1]
                a = find_all(fn, path)
                if not a:
                    urllib.request.urlretrieve(link, fn)
        else:
            if verbose:
                print("No spectra for %s found on TNS."%name)
            else:
                continue



from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.ipac.ned import Ned
import re
import numpy as np
from astro_ghost.PS1QueryFunctions import find_all

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

def getNEDSpectra(df, path, verbose=False):
    """Downloads NED spectra for the host galaxy, if it exists.

    :param df: Dataframe containing the associated transients and their host galaxies in PS1.
    :type df: Pandas DataFrame
    :param path: Filepath where NED spectra should be saved.
    :type path: str
    :param verbose: Whether to print relevant debugging information.
    :type verbose: boolean
    """

    hostNames = np.array(df.dropna(subset=['NED_name'])['NED_name'])
    transientNames = np.array(df.dropna(subset=['NED_name'])['TransientName'])
    for j in np.arange(len(hostNames)):
        try:
            spectra = Ned.get_spectra(hostNames[j])
        except:
            continue
        if spectra:
            for i in np.arange(len(spectra)):
                fn = re.sub(r"^\s+", "", hostNames[j])
                a = find_all("%s_%.02i.fits"%(fn, i), path)
                sp = spectra[i]
                if not a:
                    if verbose:
                        print("Saving spectrum %i for %s, host of %s."%(i, hostNames[j], transientNames[j]))
                    try:
                        sp.writeto(path+"/%s_%.02i.fits"%(fn, i), output_verify='ignore')
                    except:
                        if verbose:
                            print("Error in saving host spectrum for %s." % transientNames[j])
                elif verbose:
                    print("host spectrum for %s already downloaded!" % transientNames[j])

def getNEDInfo(df):
    """Gets galaxy information from NED, if it exists.

    :param df: Dataframe of the PS1 objects for which to query NED.
    :type df: Pandas DataFrame
    :return: The same dataframe as input, with NED info added.
    :rtype: Pandas DataFrame
    """

    df.reset_index(inplace=True, drop=True)

    df['NED_name'] = ""
    df['NED_type'] = ""
    df["NED_vel"] = np.nan
    df["NED_redshift"] = np.nan
    df["NED_mag"] = np.nan
    df["NED_redshift_flag"] = ""

    ra = df["raMean"]
    dec = df["decMean"]

    missingCounter = 0

    for index, row in df.iterrows():
        tempRA = ra[index]
        tempDEC = dec[index]

        # create a sky coordinate to query NED
        c = SkyCoord(ra=tempRA*u.degree, dec=tempDEC*u.degree, frame='icrs')

        # execute query
        result_table = []
        tempName = ""
        tempType = ""
        tempRed = np.nan
        tempVel = np.nan
        tempMag = np.nan

        try:
            #query NED with a 2'' radius.
            result_table = Ned.query_region(c, radius=(2/3600)*u.deg, equinox='J2000.0')
            if len(result_table) > 0:
                missingCounter = 0
        except:
            missingCounter += 1
        if len(result_table) > 0:
            #if Messier or NGC object, take that, otherwise take the closest object
            result_df = result_table.to_pandas()
            if len(result_df) > 1:
                result_NGC = result_df[[x.startswith("NGC") for x in result_df['Object Name']]]
                if len(result_NGC) > 0:
                    result_df = result_NGC.copy()
                else:
                    result_M = result_df[[x.startswith("NGC") for x in result_df['Object Name']]]
                    if len(result_M) > 0:
                        result_df = result_M.copy()
            result_gal = result_df[result_df['Type'] == b'G']
            if len(result_gal) > 0:
                result_df = result_gal.copy()
            #get closest object
            result_df = result_df[result_df['Separation'] == np.min(result_df['Separation'])]

            #remove supernovae, etc
            result_df = result_df[result_df['Type'] != b'SN']

            # Note - sometimes the colon appears in names when a sub-structure of a galaxy. Get rid of it below!
            df.loc[index, 'NED_name'] = result_df['Object Name'].values[0].split(":")[0]
            df.loc[index, 'NED_type'] = result_df['Type'].values[0]
            df.loc[index, 'NED_vel'] = result_df['Velocity'].values[0]
            df.loc[index, 'NED_redshift'] = result_df['Redshift'].values[0]
            df.loc[index, 'NED_mag'] = result_df['Redshift Flag'].values[0]

        # if the method fails for many in a row, it's likely that too many queries have been made.
        if missingCounter > 5000:
            print("Locked out of NED, will have to try again later...")
            return df
    return df



from sklearn import svm
import seaborn as sns
import matplotlib.pyplot as plt
from astro_ghost.stellarLocus import *
import pickle
import importlib_resources

def separateStars_STRM(df, model_path='.', plot=False, verbose=False, starcut='gentle'):
    """Star-galaxy separation, using a random forest trained on the PS1-STRM-classified star
       and galaxy labels given in Beck et al., 2021.

    :param df: Dataframe of PS1 sources.
    :type df: Pandas DataFrame
    :param model_path: Filepath to the saved random forest model.
    :type model_path: str, optional
    :param plot: If True, shows the separated stars and galaxies in Ap - Kron vs Ap Mag space.
    :type plot: bool, optional
    :param verbose: If true, print details of the classification routine.
    :type verbose: bool, optional
    :param starcut: Labels corresponding to the classification thresholds required to classify a star as such.
        Options are 'gentle' (P>0.8), normal (P>0.5), and aggressive (P>0.3).
    :type starcut: str

    :return: A dataframe of sources classified as galaxies.
    :rtype: Pandas DataFrame
    :return: A dataframe of sources classified as stars.
    :rtype: Pandas DataFrame
    """

    #remove known galaxies and stars
    NED_stars = df[df['NED_type'].isin(['*', '**', 'WD*', '!*', '!V*', 'V*', '!Nova'])]
    NED_gals = df[df['NED_type'] == 'G']

    #label the sources we know to be stars and galaxies
    NED_stars['sourceClass'] = 1
    NED_gals['sourceClass'] = 0

    #get the remaining objects
    unsure = df[~df.index.isin(NED_stars.index)]
    unsure = unsure[~unsure.index.isin(NED_gals.index)]

    #also drop HII regions, QSOs, etc
    unsure = unsure[~unsure['NED_type'].isin(['HII', 'QSO', 'SNR', 'PN'])]

    #remove all sources with bad values for any required PS1 properties
    unsure_dropped = unsure.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    #keep track of the unknown sources with bad values - we don't want to lose them entirely
    only_na = unsure[~unsure.index.isin(unsure_dropped.index)]

    # load random forest model
    modelName = "Star_Galaxy_RealisticModel_GHOST_PS1ClassLabels.sav"
    stream = importlib_resources.files(__name__).joinpath(modelName).open("rb")
    if verbose:
        print("Loading model %s."%modelName)
    model = pickle.load(stream)

    # plot the distribution of objects before classifying
    if plot:
        sns.set_style("dark")
        sns.set_context("talk")

        plt.figure(figsize=(10,8))
        plt.plot(unsure['iApMag'], unsure['iApMag_iKronMag'], 'o', alpha=0.1)
        plt.xlabel(r"Ap Mag, $i$")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.savefig("TNS_NoClassificationInfo.pdf")

    if len(unsure_dropped) <1:
        if verbose:
            print("No sources in field with feature values, skipping star/galaxy separation...")
        return df, unsure_dropped
    test_X = np.asarray(unsure_dropped[['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag']])

    # define probability threshold for classification
    cutdict = {'normal':0.5, 'aggressive':0.3, 'gentle':0.8}

    try:
        test_y = model.predict_proba(test_X)[:,1] > cutdict[starcut]
    except:
        print("Error! I didn't understand your starcut option.")

    unsure_dropped['sourceClass'] = test_y
    test_stars = unsure_dropped[unsure_dropped['sourceClass'] == 1]
    test_gals = unsure_dropped[unsure_dropped['sourceClass'] == 0]

    # plot the distribution of classified stars and galaxies.
    if plot:
        c_gals = '#087e8b'
        c_stars = '#ffbf00'

        plt.figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(test_gals['iApMag'], test_gals['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', color=c_gals, alpha=0.5,legend=True);
        if len(test_stars) > 1:
            sns.kdeplot(test_stars['iApMag'], test_stars['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', color=c_stars, alpha=0.5);
        plt.legend()
        plt.xlim(xmin=13,xmax=23)
        plt.ylim(ymin=-1,ymax=5)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)

        plt.savefig("TNS_STRM_Stars_vs_Gals_Contours.pdf")

        plt.figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        ax = sns.distplot(test_gals['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Galaxies',kde=False,hist_kws={"color": c_gals});
        if len(test_stars) > 1:
            sns.distplot(test_stars['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Stars',kde=False,hist_kws={"color": c_stars});
        sns.distplot(unsure['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Total',kde=False,hist_kws={"histtype": "step", "linewidth": 1, "alpha": 1, "color": "black"});

        plt.xlim(-1, 3)
        plt.xlabel("Ap - Kron Mag, $i$",fontsize=16)
        plt.xlim()
        plt.ylabel("N",fontsize=16)
        plt.legend(fontsize=16)
        plt.savefig("TNS_iAp_iKronMag_Histogram.pdf")

    # Adding back in the ones we identified as galaxies from our clustering above, and the NED sources.
    df_gals = pd.concat([test_gals, only_na, NED_gals])
    df_gals.reset_index(inplace=True, drop=True)

    # Adding back in the ones we identified as stars from our clustering above, and the NED sources.
    df_stars = pd.concat([test_stars, NED_stars])
    df_stars.reset_index(inplace=True, drop=True)
    return df_gals, df_stars

def separateStars_South(df, plot=0, verbose=0, starcut='gentle'):
    """Star-galaxy separation in the southern hemisphere, using a simple
       SkyMapper_StarClass threshold cut.

    :param df: Dataframe of PS1 sources.
    :type df: Pandas DataFrame
    :param plot: If True, shows the separated stars and galaxies in Ap - Kron vs Ap Mag space.
    :type plot: bool, optional
    :param verbose: If true, print details of the classification routine.
    :type verbose: bool, optional
    :param starcut: Labels corresponding to the classification thresholds required to classify a star as such.
        Options are 'gentle' (P>0.8), normal (P>0.5), and aggressive (P>0.3).
    :type starcut: str

    :return: A dataframe of sources classified as galaxies.
    :rtype: Pandas DataFrame
    :return: A dataframe of sources classified as stars.
    :rtype: Pandas DataFrame
    """

    #remove known galaxies and stars
    NED_stars = df[df['NED_type'].isin(['*', '**', 'WD*', '!*', '!V*', 'V*', '!Nova'])]
    NED_gals = df[df['NED_type'] == 'G']

    #get the remaining objects
    unsure = df[~df.index.isin(NED_stars.index)]
    unsure = unsure[~unsure.index.isin(NED_gals.index)]

    unsure_dropped = unsure[unsure['SkyMapper_StarClass']>0]
    only_na = unsure[~unsure.index.isin(unsure_dropped.index)]

    # define probability threshold for classification
    cutdict = {'normal':0.5, 'aggressive':0.3, 'gentle':0.8}
    try:
        test_stars = unsure_dropped[unsure_dropped['SkyMapper_StarClass']> cutdict[starcut]]
    except:
        print("Error! I didn't understand your starcut option.")

    test_gals = unsure_dropped[~unsure_dropped.index.isin(test_stars.index)]

    # Adding back in the ones we identified as galaxies from our clustering above, and the NED sources.
    df_gals = pd.concat([test_gals, NED_gals])
    df_gals['sourceClass'] = 0
    df_gals.reset_index(inplace=True, drop=True)

    # Adding back in the ones we identified as stars from our clustering above, and the NED sources.
    df_stars = pd.concat([test_stars, NED_stars])
    df_stars['sourceClass'] = 1
    df_stars.reset_index(inplace=True, drop=True)
    return df_gals, df_stars



import numpy as np
import os
import pandas as pd
from astropy import units as u
import astropy.coordinates as coord

def clean_dict(dic, df, to_keep):
    """Re-writes the transient, host list dictionary such that
       no potential hosts remain that are not in our data frame
       of potential hosts (you can pass in an array,
       to_keep, of potential hosts to keep in the
       dictionary).

    :param dic: key,value pairs of transient name, list of candidate host PS1
        objIDs.
    :type dic: dictionary
    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :param to_keep: List of PS1 objIDs for candidates to keep in the dictionary
        (even if they're not in the dataframe).
    :type to_keep: array


    :return: key,value pairs of transient name, list of candidate host PS1
        objIDs (after cleaning).
    :rtype: dictionary
    """

    for name, host in dic.items():
            host = host.tolist()
            newHosts = host
            if len(host) > 0:
                for hostCandidate in np.array(host):
                    if hostCandidate not in np.array(df["objID"]) and hostCandidate not in np.array(to_keep):
                        newHosts.remove(hostCandidate)
                dic[name] = np.array(newHosts)
    return dic

def check_dict(dic, df):
    """Check to make sure that the list of all potential hosts
       in the dataframe matches all potential hosts in the
       dictionary - that the two describe the same hosts.

    :param dic: key,value pairs of transient name, list of candidate host PS1
        objIDs.
    :type dic: dictionary
    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    """

    for name, host in dic.items():
            host = host.tolist()
            newHosts = host
            if len(host) > 0:
                for hostCandidate in np.array(host):
                    if hostCandidate not in np.array(df["objID"]):
                        print("Error: {}".format(hostCandidate))
                        print("objID not found!")

def clean_dict(dic, df, to_keep=[],bestDetectionCut=False):
    """ Clean the dictionary to match the dataset.
        note that if we're cutting by bestDetection, we find that
        we cut out many true hosts. So we only overwrite the
        list of potential hosts in this case if our new list isn't
        empty.

    :param dic: key,value pairs of transient name, list of candidate host PS1
        objIDs.
    :type dic: dictionary
    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :param to_keep: List of PS1 objIDs for candidates to keep in the dictionary
        (even if they're not in the dataframe).
    :type to_keep: array
    :param bestDetectionCut: If True, bestDetection==1 was a selection cut.
        Don't remove hosts from the dictionary if this would remove all
        of a transient's potential hosts.
    :type bestDetectionCut: bool

    :return: key,value pairs of transient name, list of candidate
        host PS1 objIDs, after removing PS1 objects not in the dataframe.
    :rtype: dictionary
    """

    for name, host in dic.items():
            host = host.tolist()
            newHosts = host
            if len(host) > 0:
                for hostCandidate in np.array(host):
                    if hostCandidate not in np.array(df["objID"]) and hostCandidate not in np.array(to_keep):
                        newHosts.remove(hostCandidate)
                if ((bestDetectionCut==0) | ((bestDetectionCut ==1) & (len(np.array(newHosts) != 0)))):
                    dic[name] = np.unique(newHosts)
                else:
                    dic[name] = np.unique(dic[name])
    return dic

def clean_df_from_dict(dic, df):
    """Remove sources from PS1 object DataFrame if not in the
       dictionary matching transients to their candidate hosts.


    :param dic: key,value pairs of transient name, list of candidate host PS1
        objIDs.
    :type dic: dictionary
    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :return: PS1 properties, after removing sources.
    :rtype: Pandas DataFrame
    """

    allHosts = np.array([])
    for name, host in dic.items():
        if host.size>0:
            try:
                allHosts = np.concatenate((allHosts, np.array(host)),axis=0)
            except:
                continue
    allHosts = np.unique(allHosts)
    for index, row in df.iterrows():
        if (row['objID'] not in allHosts):
            df.drop(index, inplace=True)
    return df

def removePS1Duplicates(df):
    """Because there are many duplicate entries in PS1 host table, follow this hierarchy for
        prioritizing which ones to keep:
        1. If duplicate, remove non-primary detections
        2. If still duplicate, remove NANs in yKronFlux, yskyErr, and yExtNSigma
        3. If still duplicate, take the value with the smallest yKronFluxErr

    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :return: PS1 properties, after removing duplicates.
    :rtype: Pandas DataFrame
    """

    df.replace(-999.0,np.nan, inplace=True)
    new_df = []
    for hostCandidate in np.unique(df["objID"]):
        hostFrame = df[df["objID"] == hostCandidate]
        if len(hostFrame) > 1:
            newhostFrame = hostFrame[hostFrame["primaryDetection"] == 1]
            if len(newhostFrame) > 1:
                newhostFrame = newhostFrame[~np.isnan(newhostFrame["rApMag"])]
            if len(newhostFrame) > 1:
                newhostFrame = newhostFrame[newhostFrame["rApMagErr"] == np.min(newhostFrame["rApMagErr"])]
            if len(newhostFrame) > 0:
                new_df.append(newhostFrame)
            else:
                new_df.append(hostFrame)
        else:
            new_df.append(hostFrame)
    if len(new_df) > 0:
        df = pd.concat(new_df)
    return df

def getColors(df):
    """Calulate observer-frame colors for PS1 sources, and make some cuts
       from bad photometry.

    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :return: PS1 properties, with color and photometry cuts.
    :rtype: Pandas DataFrame
    """

    df.replace(-999, np.nan, inplace=True)
    df.replace(999, np.nan, inplace=True)

    # create color attributes for all hosts
    df["g-r"]= df["gApMag"] - df["rApMag"]
    df["r-i"]= df["rApMag"] - df["iApMag"]
    df["i-z"] = df["iApMag"] - df["zApMag"]
#    df["g-i"] = df["gApMag"] - df["iApMag"]
    df["z-y"] = df["zApMag"] - df["yApMag"]

    df['g-rErr'] = np.sqrt(df['gApMagErr']**2 + df['rApMagErr']**2)
    df['r-iErr'] = np.sqrt(df['rApMagErr']**2 + df['iApMagErr']**2)
    df['i-zErr'] = np.sqrt(df['iApMagErr']**2 + df['zApMagErr']**2)
    df['z-yErr'] = np.sqrt(df['zApMagErr']**2 + df['yApMagErr']**2)

    # To be sure we're getting physical colors
    for col in ['g-r', 'r-i', 'i-z', 'z-y']:
        df.loc[np.abs(df[col]) > 100, col] = np.nan

    # and PSF - Kron mag "morphology" information
    for band in 'grizy':
        # to be sure we're getting physical mags
        col = '%sApMag_%sKronMag'%(band, band)
        df[col] = df["%sApMag"%band] - df["%sKronMag"%band]
        df.loc[np.abs(df[col]) > 100, col] = np.nan
        col = '%sApMag'%band
        df.loc[np.abs(df[col]) > 100, col] = np.nan
    return df

def makeCuts(df,cuts=[],dict=""):
    """Make a series of quality cuts on the candidate host galaxies in the dataframe.

    :param df: PS1 properties for candidate hosts.
    :type df: Pandas DataFrame
    :param cuts: List of cuts to apply. Options are:
        'n' - remove objects without at least 10 detections.
        'quality'- remove objects with PS1 qualityFlag > 128 (suggesting bad photometry).
        'coords' - remove objects with missing position information.
        'mag' - remove objects with missing photometry (aperture magnitudes).
        'primary' - remove objects with primaryDetection = 0.
        'best' - remove objects with bestDetection = 0.
        'duplicate' - remove all completely duplicated rows.
    :type cuts: array
    :param dic: key,value pairs of transient name, list of candidate host PS1
        objIDs.
    :type dic: dictionary

    :return: PS1 properties, with quality cuts applied.
    :rtype: Pandas DataFrame
    """

    for cut in cuts:
        if cut == "n":
            df = df[df['nDetections'] >= 10]
        elif cut == "quality":
            df = df[df["qualityFlag"] < 128]
        elif cut == "coords":
            df = df.dropna(subset=['raMean', 'decMean'])
        elif cut == "mag":
            for band in 'grizy':
                df = df[pd.notnull(df['%sApMag'%band])]
        elif cut =="primary":
            if dict != "":
                df_mod = df[df["primaryDetection"] == 1]
                clean_dict(dict, df_mod, np.array([]),bestDetectionCut=1)
                clean_df_from_dict(dict, df)
            else:
                print("Can't make a primary detection cut - no dict provided!")
                return
        elif cut == "best":
            # subset by best detection, clean the dictionary, but include a flag that only
            # cleans the dict entry if there was a best detection in the field (else leave unchanged!)
            if dict != "":
                df_mod = df[df['bestDetection'] == 1]
                clean_dict(dict, df_mod, np.array([]),bestDetectionCut=1)
                clean_df_from_dict(dict, df)
            else:
                print("Can't make a best detection cut - no dict provided!")
                return
        elif cut=="duplicate":
            df = df.drop_duplicates(subset=['objID'])
        else:
            print("I didn't understand your cut!")
            return
    df.reset_index(drop=True, inplace=True)
    return df



import os
import pytest
import sys
from astro_ghost.PS1QueryFunctions import *
from astro_ghost.TNSQueryFunctions import *
from astro_ghost.NEDQueryFunctions import *
from astro_ghost.ghostHelperFunctions import *
from astro_ghost.starSeparation import *
from astro_ghost.stellarLocus import *
from astropy.coordinates import SkyCoord
from astropy import units as u
import pandas as pd
from datetime import datetime

#we want to include print statements so we know what the algorithm is doing
verbose = 1

def test_getGHOST():
    #Download the GHOST database.
    #note: real=False creates an empty database, which
    #allows you to use the association methods without
    #needing to download the full database first

    getGHOST(real=True, verbose=verbose, clobber=True)
    #test that it got the ghost database
    df = fullData()
    # GHOST has the correct match for at least NGC 2997
    assert df.loc[df['TransientName'] == 'SN2003jg', 'NED_name'].values[0] == 'NGC 2997'

def test_NED():
    # test our ability to snag a galaxy name from NED, using the coordinates of NGC 4321
    df = pd.DataFrame({'objID':23412341234, 'raMean':[185.7288750], 'decMean':[15.82230]})
    df = getNEDInfo(df)
    assert df['NED_name'].values[0] == 'NGC 4321'

def test_associate():
    #create a list of the supernova names, their skycoords, and their classes (these three are from TNS)
    snName = ['SN 2012dt', 'SN 1998bn', 'SN 1957B']

    snCoord = [SkyCoord(14.162*u.deg, -9.90253*u.deg, frame='icrs'), \
            SkyCoord(187.32867*u.deg, -23.16367*u.deg, frame='icrs'), \
            SkyCoord(186.26125*u.deg, +12.899444*u.deg, frame='icrs')]

    snClass = ['SN IIP', 'SN', 'SN Ia']

    # run the association algorithm with the DLR method!
    hosts = getTransientHosts(snName, snCoord, snClass, verbose=verbose, starcut='gentle', ascentMatch=False)

    correctHosts = [SkyCoord(14.1777425*u.deg, -9.9138756*u.deg, frame='icrs'),
                    SkyCoord(187.3380517*u.deg, -23.1666716*u.deg, frame='icrs'),
                    SkyCoord(186.2655971*u.deg, 12.8869831*u.deg, frame='icrs')]

    sep = []
    for i in np.arange(len(correctHosts)):
       c1 = correctHosts[i]
       c2 = SkyCoord(hosts['TransientRA'].values[i]*u.deg, hosts['TransientDEC'].values[i]*u.deg, frame='icrs')
       sep.append(c2.separation(c2).arcsec)

    #consider a success if the three hosts were found to a 1'' precision
    assert np.nanmax(sep) < 1


def test_starSeparation():
    #classify a few galaxies, classify a few stars
    sourceType = ['galaxy',  'galaxy', 'star', 'star']
    ra = [186.7154417, 31.0672500, 191.0597417, 17.8279833]
    dec = [9.1342306, 20.8474806, 12.3629917, 33.2604472]
    sourceSet = []
    for i in np.arange(len(ra)):
        a = ps1cone(ra[i], dec[i], 1/3600)
        a = ascii.read(a)
        sourceSet.append(a.to_pandas().iloc[[0]])
    sourceDF = pd.concat(sourceSet, ignore_index=True)
    sourceDF['trueSourceClass'] = sourceType
    sourceDF = getColors(sourceDF)
    sourceDF = calc_7DCD(sourceDF)
    sourceDF = getNEDInfo(sourceDF)
    gals, stars = separateStars_STRM(sourceDF)

    #True is stars, False is gals. Passes if all gals are False and all stars are True!
    assert (np.nansum(gals['sourceClass'] == False) + np.nansum(stars['sourceClass'])) == len(sourceDF)

def test_plotLocus():
    #plot a subset of GHOST data compared to the tonry stellar locus
    GHOST = fullData()
    plotLocus(GHOST.iloc[0:500], color=False, save=True, type="Gals", timestamp="")
    assert len(glob.glob(os.path.join(os.getcwd(), 'PS1_Gals_StellarLocus_.pdf'))) > 0

def test_skymapper_associate():
    snName = ['AT2023hri']
    snCoord = [SkyCoord(288.5964917*u.deg,  -54.5636583*u.deg, frame='icrs')]
    snClass = ['AT']

    # run the association algorithm with the DLR method!
    hosts = getTransientHosts(snName, snCoord, snClass, verbose=verbose, starcut='gentle', ascentMatch=False)
    assert hosts['NED_name'].values[0] == 'ESO 184- G 042'

def test_resolve():
    coords = resolve('M100')
    c1 = SkyCoord(coords[0]*u.deg, coords[1]*u.deg, frame='icrs')
    c2 = SkyCoord(185.7288750*u.deg, 15.82230*u.deg, frame='icrs')
    assert c2.separation(c2).arcsec < 0.1



import numpy as np
from PIL import Image
from io import BytesIO
import os
import pandas as pd
import sys
import re
import json
import mastcasjobs
import requests
from datetime import datetime
try: # Python 3.x
    from urllib.parse import quote as urlencode
    from urllib.request import urlretrieve
    import http.client as httplib
except ImportError:  # Python 2.x
    from urllib import pathname2url as urlencode
    from urllib import urlretrieve
    import httplib
from collections import OrderedDict
from os import listdir
from os.path import isfile, join
from astropy import utils, io, convolution, wcs
from astropy import units as u
from astropy.visualization import make_lupton_rgb,PercentileInterval, AsinhStretch
from astropy.coordinates import name_resolve, Angle
from astropy.io import fits, ascii
from astropy.table import Table
from pyvo.dal import sia
import pickle
from io import BytesIO

from astropy.coordinates import SkyCoord
from astroquery.vo_conesearch import ConeSearch
from astroquery.vo_conesearch import vos_catalog
vos_catalog.list_catalogs("conesearch_good")
from warnings import simplefilter

# could absolutely be more efficient, but filter out warnings for now
simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

# set a few environmental variables that we'll need to query PS2 for more accurate galaxy sizes.
if "CASJOBS_USERID" not in os.environ:
    os.environ['CASJOBS_USERID'] = 'ghostbot'
    os.environ['CASJOBS_PW'] = 'ghostbot'

def getAllPostageStamps(df, tempSize, path=".", verbose=False):
    """Loops through a pandas dataframe and saves PS1 stacked color images of all
       host galaxies.

    :param df: Dataframe of PS1 sources.
    :type df: Pandas DataFrame
    :param tempSize: The downloaded image will be tempSize x tempSize pixels.
    :type tempSize: int
    :param path: Filepath where images should be saved.
    :type path: str
    :param verbose: If true, The progress of the image downloads is printed.
    :type verbose: bool
    """

    for i in np.arange(len(df["raMean"])):
            tempRA = df.loc[i, 'raMean']
            tempDEC = df.loc[i, 'decMean']
            tempName = df.loc[i, 'TransientName']
            a = find_all(path+"/%s.png" % tempName, path)
            if not a:
                try:
                    img = getcolorim(tempRA, tempDEC, size=tempSize, filters="grizy", format="png")
                    img.save(path+"/%s.png" % tempName)
                    if verbose:
                        print("Saving postage stamp for the host of %s."% tempName)
                except:
                    continue

def get_hosts(path, transient_fn, fn_Host, rad):
    """Wrapper function for getting candidate host galaxies in PS1 for dec>-30 deg and
       in Skymapper for dec<-30 deg.

    :param path: Filepath where csv of candidate hosts should be saved.
    :type path: str
    :param transient_fn: Filename of csv containing the transients to associate (and their coordinates).
    :type transient_fn: str
    :param fn_Host: Filename of csv containing candidate host galaxy properties.
    :type fn_Host: str
    :param rad: Search radius of the algorithm, in arcseconds.
    :type rad: float
    :return: Dataframe of all candidate host galaxies.
    :rtype: Pandas DataFrame
    """

    transient_df = pd.read_csv(path+"/"+transient_fn)
    now = datetime.now()
    dict_fn = fn_Host.replace(".csv", "") + ".p"

    tempDEC = Angle(transient_df['DEC'], unit=u.deg)
    tempDEC = tempDEC.deg

    # distinguish between PS1 sources and Skymapper sources.
    df_North = transient_df[(tempDEC > -30)].reset_index()
    df_South = transient_df[(tempDEC <= -30)].reset_index()

    # get southern-hemisphere sources
    append=0
    if len(df_South) > 0:
        print("Finding southern sources with SkyMapper...")
        find_host_info_SH(df_South, fn_Host, dict_fn, path, rad)
        append=1

    # get northern-hemisphere sources
    if len(df_North) > 0:
        print("Finding northern sources with Pan-starrs...")
        find_host_info_PS1(df_North, fn_Host, dict_fn, path, rad, append=append)

    # load the saved csv and remove duplicates, then return
    host_df = pd.read_csv(path+"/"+fn_Host)
    host_df = host_df.drop_duplicates()
    host_df.to_csv(path+"/"+fn_Host[:-4]+"_cleaned.csv", index=False)
    return host_df

def find_all(name, path):
    """Crawls through a directory and all its sub-directories looking for a file matching
       \\'name\\'. If found, it is returned.

    :param name: The filename for which to search.
    :type name: str
    :param path: The directory to search.
    :type path: str
    :return: The list of absolute paths to all files called \\'name\\' in \\'path\\'.
    :rtype: list
    """

    result = []
    for root, dirs, files in os.walk(path):
        if name in files:
            result.append(os.path.join(root, name))
    return result

def getimages(ra,dec,size=240,filters="grizy", type='stack'):
    """Query ps1filenames.py service to get a list of images.

    :param ra: Right ascension of position, in degrees.
    :type ra: float
    :param dec: Declination of position, in degrees.
    :type dec: float
    :param size: The image size in pixels (0.25 arcsec/pixel)
    :type size: int
    :param filters: A string with the filters to include
    :type filters: str
    :return: The results of the search for relevant images.
    :rtype: Astropy Table
    """

    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = ("{service}?ra={ra}&dec={dec}&size={size}&format=fits"
           "&filters={filters}&type={type}").format(**locals())
    table = Table.read(url, format='ascii')
    return table


def geturl(ra, dec, size=240, output_size=None, filters="grizy", format="jpg", color=False, type='stack'):
    """Get the URL for images in the table.

    :param ra: Right ascension of position, in degrees.
    :type ra: float
    :param dec: Declination of position, in degrees.
    :type dec: float
    :param size: The extracted image size in pixels (0.25 arcsec/pixel)
    :type size: int
    :param output_size: output (display) image size in pixels (default = size).
        The output_size has no effect for fits format images.
    :type output_size: int
    :param filters: The string with filters to include.
    :type filters: str
    :param format: The data format (options are \\"jpg\\", \\"png" or \\"fits\\").
    :type format: str
    :param color: If True, creates a color image (only for jpg or png format).
        If False, return a list of URLs for single-filter grayscale images.
    :type color: bool, optional
    :return: The url for the image to download.
    :rtype: str
    """

    if color and format == "fits":
        raise ValueError("color images are available only for jpg or png formats")
    if format not in ("jpg","png","fits"):
        raise ValueError("format must be one of jpg, png, fits")
    table = getimages(ra,dec,size=size,filters=filters, type=type)
    url = ("https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
           "ra={ra}&dec={dec}&size={size}&format={format}").format(**locals())
    if output_size:
        url = url + "&output_size={}".format(output_size)

    # sort filters from red to blue
    flist = ["yzirg".find(x) for x in table['filter']]
    table = table[np.argsort(flist)]
    if color:
        if len(table) > 3:
            # pick 3 filters
            table = table[[0,len(table)//2,len(table)-1]]
        for i, param in enumerate(["red","green","blue"]):
            url = url + "&{}={}".format(param,table['filename'][i])
    else:
        urlbase = url + "&red="
        url = []
        for filename in table['filename']:
            url.append(urlbase+filename)
    return url


def getcolorim(ra, dec, size=240, output_size=None, filters="grizy", format="jpg"):
    """Get a PS1 color image at a sky position.

    :param ra: Right ascension of position, in degrees.
    :type ra: float
    :param dec: Declination of position, in degrees.
    :type dec: float
    :param size: The extracted image size in pixels (0.25 arcsec/pixel)
    :type size: int
    :param output_size: output (display) image size in pixels (default = size).
        The output_size has no effect for fits format images.
    :type output_size: int
    :param filters: The string with filters to include.
    :type filters: str
    :param format: The data format (options are \\'jpg\\', \\'png\\' or \\'fits\\').
    :type format: str
    :return: The image.
    :rtype: PIL Image
    """

    if format not in ("jpg","png"):
        raise ValueError("format must be jpg or png")
    url = geturl(ra,dec,size=size,filters=filters,output_size=output_size,format=format,color=True, type='stack')
    r = requests.get(url)
    im = Image.open(BytesIO(r.content))
    return im

def get_PS1_type(path, ra, dec, size, band, type):
    """Download and save PS1 imaging data in a given band of a given type.

    :param path: filepath where the image will be saved.
    :type path: str
    :param ra: Right ascension of position, in degrees.
    :type ra: float
    :param dec: Declination of position, in degrees.
    :type dec: float
    :param size: The extracted image size in pixels (0.25 arcsec/pixel)
    :type size: int
    :param band: The PS1 band.
    :type band: str
    :param type: The type of imaging data to obtain. Options are given below for the PS1 stack.
        \\'stack.mask\\' - indicate which pixels in the stack are good and which are bad
        \\'stack.wt\\' - stack variance images
        \\'stack.num\\' - contain the number of warps with valid data which contributed to each pixel
        \\'stack.exp\\' - contain the exposure time in seconds which contributed to each pixel
        \\'stack.expw\\' - weighted exposure time maps
        See more information at https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images.
    :type type: str
    """

    fitsurl = geturl(ra, dec, size=size, filters="{}".format(band), format="fits", type=type)
    fh = fits.open(fitsurl[0])
    fh.writeto(path + '/PS1_ra={}_dec={}_{}arcsec_{}_{}.fits'.format(ra, dec, int(size*0.25), band, type))

def get_PS1_Pic(path, objID, ra, dec, size, band, safe=False):
    """Downloads PS1 picture (in fits) centered at a given location.

    :param path: The filepath where the fits file will be saved.
    :type path: str
    :param objID: The PS1 objID of the object of interest (to save as filename).
    :type objID: int
    :param ra: Right ascension of position, in degrees.
    :type ra: float
    :param dec: Declination of position, in degrees.
    :type dec: float
    :param size: The extracted image size in pixels (0.25 arcsec/pixel)
    :type size: int
    :param band: The PS1 band.
    :type band: str
    :param safe: If True, include the objID of the object of interest in the filename
        (useful when saving multiple files at comparable positions).
    :type safe: bool, optional
    """

    fitsurl = geturl(ra, dec, size=size, filters="{}".format(band), format="fits")
    fh = fits.open(fitsurl[0])
    if safe==True:
        fh.writeto(path + '/PS1_{}_{}arcsec_{}.fits'.format(objID, int(size*0.25), band))
    else:
        fh.writeto(path + '/PS1_ra={}_dec={}_{}arcsec_{}.fits'.format(ra, dec, int(size*0.25), band))

def ps1metadata(table="mean",release="dr1",baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs"):
    """Return metadata for the specified catalog and table.

    :param table: Table type. Can be \\'mean\\', \\'stack\\', or \\'detection\\'
    :type table: str
    :param release: The Pan-STARRS data release. Can be \\'dr1' or \\'dr2\\'.
    :type release: str
    :param baseurl: The base URL for the request
    :type baseurl: str
    :return: The table containing the metadata, with columns name, type, and description.
    :rtype: Astropy Table
    """

    checklegal(table,release)
    url = "{baseurl}/{release}/{table}/metadata".format(**locals())
    r = requests.get(url)
    r.raise_for_status()
    v = r.json()

    # convert to astropy table
    tab = Table(rows=[(x['name'],x['type'],x['description']) for x in v],
               names=('name','type','description'))
    return tab


def mastQuery(request):
    """Perform a MAST query.

    :param request: The MAST request json object.
    :type request: dictionary
    :return: The response HTTP headers.
    :rtype: HTTP header
    :return: The data obtained.
    :rtype: str
    """

    server='mast.stsci.edu'

    # Grab Python Version
    version = ".".join(map(str, sys.version_info[:3]))

    # Create Http Header Variables
    headers = {"Content-type": "application/x-www-form-urlencoded",
               "Accept": "text/plain",
               "User-agent":"python-requests/"+version}

    # Encoding the request as a json string
    requestString = json.dumps(request)
    requestString = urlencode(requestString)

    # opening the https connection
    conn = httplib.HTTPSConnection(server)

    # Making the query
    conn.request("POST", "/api/v0/invoke", "request="+requestString, headers)

    # Getting the response
    resp = conn.getresponse()
    head = resp.getheaders()
    content = resp.read().decode('utf-8')

    # Close the https connection
    conn.close()

    return head,content

def resolve(name):
    """Get the RA and Dec for an object using the MAST name resolver.

    :param name: Name of the object of interest.
    :type name: str
    :return: Position of resolved object, in degrees.
    :rtype: tuple
    """

    resolverRequest = {'service':'Mast.Name.Lookup',
                       'params':{'input':name,
                                 'format':'json'
                                },
                      }
    headers,resolvedObjectString = mastQuery(resolverRequest)
    resolvedObject = json.loads(resolvedObjectString)

    # The resolver returns a variety of information about the resolved object,
    # however for our purposes all we need are the RA and Dec
    try:
        objRa = resolvedObject['resolvedCoordinate'][0]['ra']
        objDec = resolvedObject['resolvedCoordinate'][0]['decl']
    except IndexError as e:
        raise ValueError("Unknown object '{}'".format(name))
    return (objRa, objDec)

def checklegal(table,release):
    """Checks if this combination of table and release is acceptable.
       Raises a VelueError exception if there is problem.

    :param table: Table type. Can be \\'mean\\', \\'stack\\', or \\'detection\\'
    :type table: str
    :param release: The Pan-STARRS data release. Can be \\'dr1\\' or \\'dr2\\'.
    :type release: str
    :raises ValueError: Raises error if table and release combination are invalid.
    """

    releaselist = ("dr1", "dr2")
    if release not in ("dr1","dr2"):
        raise ValueError("Bad value for release (must be one of {})".format(', '.join(releaselist)))
    if release=="dr1":
        tablelist = ("mean", "stack")
    else:
        tablelist = ("mean", "stack", "detection")
    if table not in tablelist:
        raise ValueError("Bad value for table (for {} must be one of {})".format(release, ", ".join(tablelist)))

def ps1search(table="mean",release="dr1",format="csv",columns=None,baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs", verbose=False,**kw):
    """Do a general search of the PS1 catalog (possibly without ra/dec/radius).

    :param table: Table type. Can be \\'mean\\', \\'stack\\', or \\'detection\\'
    :type table: str
    :param release: The Pan-STARRS data release. Can be \\'dr1\\' or 'dr2\\'.
    :type release: str
    :param format: Can be \\'csv\\', \\'votable\\', or \\'json\\'.
    :type format: str
    :param columns: Column names to include (None means use defaults).
    :type columns: str
    :param baseurl: Base URL for the request.
    :type baseurl: str
    :param verbose: If true, print info about request.
    :type verbose: bool
    :param \\*\\*kw: Other parameters (e.g., \\'nDetections.min\\':2).  Note that this is required!
    :type \\*\\*kw: dictionary
    :return: Result of PS1 query, in \\'csv\\', \\'votable\\', or \\'json\\' format.
    :rtype: Same as \\'format\\'
    """

    data = kw.copy()
    if not data:
        raise ValueError("You must specify some parameters for search")
    checklegal(table,release)
    if format not in ("csv","votable","json"):
        raise ValueError("Bad value for format")
    url = "{baseurl}/{release}/{table}.{format}".format(**locals())
    if columns:
        # check that column values are legal
        # create a dictionary to speed this up
        dcols = {}
        for col in ps1metadata(table,release)['name']:
            dcols[col.lower()] = 1
        badcols = []
        for col in columns:
            if col.lower().strip() not in dcols:
                badcols.append(col)
        if badcols:
            raise ValueError('Some columns not found in table: {}'.format(', '.join(badcols)))
        # two different ways to specify a list of column values in the API
        data['columns'] = '[{}]'.format(','.join(columns))

    # either get or post works
    r = requests.get(url, params=data)

    if verbose:
        print(r.url)
    r.raise_for_status()
    if format == "json":
        return r.json()
    else:
        return r.text


def ps1cone(ra,dec,radius,table="stack",release="dr1",format="csv",columns=None,baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs", verbose=False,**kw):
    """Do a cone search of the PS1 catalog. Note that this is just a thin wrapper for the function \\'ps1search\\'.

    :param ra: Right ascension of central coordinate, in J200 degrees.
    :type ra: float
    :param dec: Declination of central coordinate, in J2000 degrees.
    :type dec: float
    :param radius: Search radius, in degrees (<= 0.5 degrees)
    :type radius: float
    :param table: Table type. Can be \\'mean\\', \\'stack\\', or \\'detection\\'
    :type table: str
    :param release: The Pan-STARRS data release. Can be \\'dr1\\' or \\'dr2\\'.
    :type release: str
    :param format: Can be \\'csv\\', \\'votable\\', or \\'json\\'.
    :type format: str
    :param columns: Column names to include (None means use defaults).
    :type columns: str
    :param baseurl: Base URL for the request.
    :type baseurl: str
    :param verbose: If true, print info about request.
    :type verbose: bool
    :param \\*\\*kw: Other parameters (e.g., \\'nDetections.min\\':2).  Note that this is required!
    :type \\*\\*kw: dictionary
    :return: Result of PS1 query, in \\'csv\\', \\'votable\\', or \\'json\\' format.
    :rtype: Same as \\'format\\'
    """

    data = kw.copy()
    data['ra'] = ra
    data['dec'] = dec
    data['radius'] = radius
    return ps1search(table=table,release=release,format=format,columns=columns,
                    baseurl=baseurl, verbose=verbose, **data)

def ps1metadata(table="mean", release="dr1", baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs"):
    """Return metadata for the specified catalog and table. Snagged from the
       wonderful API at https://ps1images.stsci.edu/ps1_dr2_api.html.

    :param table: Table type. Can be \\'mean\\', \\'stack\\', or \\'detection\\'
    :type table: str
    :param release: The Pan-STARRS data release. Can be \\'dr1\\' or \\'dr2\\'.
    :type release: str
    :param baseurl: Base URL for the request.
    :type baseurl: str
    :return: Table with columns name, type, description.
    :rtype: Astropy Table
    """

    checklegal(table,release)
    url = f"{baseurl}/{release}/{table}/metadata"
    r = requests.get(url)
    r.raise_for_status()
    v = r.json()

    # convert to astropy table
    tab = Table(rows=[(x['name'],x['type'],x['description']) for x in v],
               names=('name','type','description'))
    return tab

def find_host_info_PS1(df, fn, dict_fn, path, rad, append=0):
    """Querying PS1 for all objects within rad arcsec of each SN.

    :param df: Dataframe of transient information (name and coordinates).
    :type df: Pandas DataFrame
    :param fn: Filename of PS1 candidate host dataframe.
    :type fn: str
    :param dict_fn: The filename of the dictionary to keep track of transient-candidate host matches.
        Keys are transient names, values are lists containing objIDs of all host candidates.
    :type dict_fn: dictionary
    :param path: The filepath where df will be saved.
    :type path: str
    :param rad: The search radius, in arcsec.
    :type rad: float
    :param append: If True, append results to fn. If False, create a new file.
    :type append: bool, optional
    """

    i = 0

    # The dictionary to map SN names to nearby obj IDs in PS1
    SN_Host_PS1 = {}

    #a running list of all PS1 results
    PS1_queries = []
    for j, row in enumerate(df.itertuples(), 1):
            if ":" in str(row.RA):
                tempRA = Angle(row.RA, unit=u.hourangle)
            else:
                tempRA = Angle(row.RA, unit=u.deg)
            tempDEC = Angle(row.DEC, unit=u.deg)
            a = ps1cone(tempRA.degree,tempDEC.degree,rad/3600,table="stack",release="dr1",format="csv",columns=None,baseurl="https://catalogs.mast.stsci.edu/api/v0.1/panstarrs", verbose=False)
            if a:
                a = ascii.read(a)
                a = a.to_pandas()
                PS1_queries.append(a)
                SN_Host_PS1[row.Name] = np.array(a['objID'])
            else:
                SN_Host_PS1[row.Name] = np.array([])

            # Print status messages every 10 lines
            if j%10 == 0:
                print("Processed {} of {} lines!".format(j, len(df.Name)))

            # Print every query to a file. This was done in order
            # to prevent the code crashing after processing 99% of the data
            # frame and losing everything. This allows for duplicates, though,
            # so they should be removed before the file is used again.
            if (len(PS1_queries) > 0):
                PS1_hosts = pd.concat(PS1_queries)
                PS1_hosts = PS1_hosts.drop_duplicates()

                #match up rows to skymapper cols to join into a single dataframe
                newCols = np.array(['SkyMapper_StarClass', 'gelong','g_a','g_b','g_pa',
                'relong','r_a','r_b','r_pa',
                'ielong','i_a','i_b','i_pa',
                'zelong','z_a','z_b','z_pa'])
                for col in newCols:
                    PS1_hosts[col] = np.nan
                PS1_queries = []
                if not append:
                    PS1_hosts.to_csv(path+fn, header=True, index=False)
                    i = 1
                    append = True
                else:
                    PS1_hosts.to_csv(path+"/"+fn, mode='a+', header=False, index=False)
            else:
                print("No potential hosts found for this object...")

            # Save host info
            if not os.path.exists(path+ '/dictionaries/'):
            	os.makedirs(path+'/dictionaries/')
            option = "wb"
            if append:
                option = "ab"
            with open(path+"/dictionaries/" + dict_fn, option) as fp:
                pickle.dump(SN_Host_PS1, fp, protocol=pickle.HIGHEST_PROTOCOL)

def find_host_info_SH(df, fn, dict_fn, path, rad):
    """VO Cone Search for all objects within rad arcsec of SNe (for Southern-Hemisphere (SH) objects).

    :param df: Dataframe of transient information (name and coordinates).
    :type df: Pandas DataFrame
    :param fn: Filename of PS1 candidate host dataframe.
    :type fn: str
    :param dict_fn: The filename of the dictionary to keep track of transient-candidate host matches.
        Keys are transient names, values are lists containing objIDs of all host candidates.
    :type dict_fn: dictionary
    :param path: The filepath where df will be saved.
    :type path: str
    :param rad: The search radius, in arcsec.
    :type rad: float
    :param append: If True, append results to fn. If False, create a new file.
    :type append: bool, optional
    """

    i = 0
    SN_Host_SH = {}
    SH_queries = []
    pd.options.mode.chained_assignment = None
    for j, row in df.iterrows():
            if ":" in str(row.RA):
                tempRA = Angle(row.RA, unit=u.hourangle)
            else:
                tempRA = Angle(row.RA, unit=u.deg)
            tempDEC = Angle(row.DEC, unit=u.deg)
            a = pd.DataFrame({})
            a = southernSearch(tempRA.degree,tempDEC.degree, rad)
            if len(a)>0:
                SH_queries.append(a)
                SN_Host_SH[row.Name] = np.array(a['objID'])
            else:
                SN_Host_SH[row.Name] = np.array([])

            # Print status messages every 10 lines
            if j%10 == 0:
                print("Processed {} of {} lines!".format(j, len(df.Name)))

            # Print every query to a file Note: this was done in order
            # to prevent the code crashing after processing 99% of the data
            # frame and losing everything. This allows for duplicates though,
            # so they should be removed before the file is used again
            if (len(SH_queries) > 0):
                SH_hosts = pd.concat(SH_queries)
                SH_hosts = SH_hosts.drop_duplicates()
                SH_queries = []
                if i == 0:
                    SH_hosts.to_csv(path+"/"+fn, header=True, index=False)
                    i = 1
                else:
                    SH_hosts.to_csv(path+"/"+fn, mode='a+', header=False, index=False)
            else:
                print("No potential hosts found for this object...")

            # Save host info
            if not os.path.exists(path+ '/dictionaries/'):
            	os.makedirs(path+'/dictionaries/')
            with open(path+"/dictionaries/" + dict_fn, 'wb') as fp:
                pickle.dump(SN_Host_SH, fp, protocol=pickle.HIGHEST_PROTOCOL)
    pd.options.mode.chained_assignment = 'warn'

def southernSearch(ra, dec, rad):
    """Conducts a cone search for Skymapper objects at a given position.

    :param ra: Right ascension of central coordinate, in J200 degrees.
    :type ra: float
    :param dec: Declination of central coordinate, in J2000 degrees.
    :type dec: float
    :param radius: Search radius, in degrees (<= 0.5 degrees)
    :type radius: float
    :return: Dataframe of Skymapper objects formatted to be joined with PS1 sources.
    :rtype: Pandas DataFrame.
    """

    searchCoord = SkyCoord(ra*u.deg, dec*u.deg, frame='icrs')
    responseMain = requests.get("http://skymapper.anu.edu.au/sm-cone/public/query?CATALOG=dr2.master&RA=%.5f&DEC=%.5f&SR=%.5f&RESPONSEFORMAT=CSV&VERB=3" %(ra, dec, (rad/3600)))
    responsePhot = requests.get("http://skymapper.anu.edu.au/sm-cone/public/query?CATALOG=dr2.photometry&RA=%.5f&DEC=%.5f&SR=%.5f&RESPONSEFORMAT=CSV&VERB=3" %(ra, dec, (rad/3600)))

    dfMain = pd.read_csv(BytesIO(responseMain.content))
    dfPhot = pd.read_csv(BytesIO(responsePhot.content))

    filt_dfs = []
    for filter in 'griz':

        #add the photometry columns
        tempDF = dfPhot[dfPhot['filter']==filter]
        if len(tempDF) <1:
            tempDF.append(pd.Series(), ignore_index=True) #add dummy row for the sake of not crashing
        for col in tempDF.columns.values:
            if col != 'object_id':
                tempDF[filter + col] = tempDF[col]
                del tempDF[col]

        # take the column with the smallest uncertainty in the measured semi-major axis -
        # this is what we'll use to
        # calculate DLR later!
        tempDF = tempDF.loc[tempDF.groupby("object_id")[filter + 'e_a'].idxmin()]
        filt_dfs.append(tempDF)

    test = filt_dfs[0]

    for i in np.arange(1, len(filt_dfs)):
        test = test.merge(filt_dfs[i], on='object_id', how='outer')

    test['object_id'] =  np.nan_to_num(test['object_id'])
    test['object_id'] = test['object_id'].astype(np.int64)

    fullDF = test.merge(dfMain, on='object_id', how='outer')

    flag_mapping = {'objID':'object_id', 'raMean':'raj2000',
        'decMean':'dej2000','gKronRad':'gradius_kron',
        'rKronRad':'rradius_kron', 'iKronRad':'iradius_kron',
        'zKronRad':'iradius_kron', 'yKronRad':np.nan,
        'gPSFMag':'g_psf', 'rPSFMag':'r_psf','iPSFMag':'i_psf',
        'zPSFMag':'z_psf','yPSFMag':np.nan,'gPSFMagErr':'e_g_psf',
        'rPSFMagErr':'e_r_psf','iPSFMagErr':'e_i_psf','zPSFMagErr':'e_z_psf',
        'yPSFMagErr':np.nan,'gKronMag':'g_petro', 'rKronMag':'r_petro',
        'iKronMag':'i_petro', 'zKronMag':'z_petro', 'yKronMag':np.nan,
        'gKronMagErr':'e_g_petro', 'rKronMagErr':'e_r_petro',
        'iKronMagErr':'e_i_petro', 'zKronMagErr':'e_z_petro',
        'yKronMagErr':np.nan, 'ng':'g_ngood', 'nr':'r_ngood', 'ni':'i_ngood',
        'nz':'z_ngood', 'graErr':'e_raj2000', 'rraErr':'e_raj2000', 'iraErr':'e_raj2000',
        'zraErr':'e_raj2000', 'gdecErr':'e_dej2000', 'rdecErr':'e_dej2000','idecErr':'e_dej2000',
        'zdecErr':'e_dej2000','l':'glon', 'b':'glat', 'gra':'gra_img', 'rra':'rra_img', 'ira':'ira_img',
        'zra':'zra_img', 'yra':'raj2000', 'gdec':'gdecl_img', 'rdec':'rdecl_img',
        'idec':'idecl_img', 'zdec':'zdecl_img', 'ydec':'dej2000', 'gKronFlux':'gflux_kron',
        'rKronFlux':'rflux_kron', 'iKronFlux':'iflux_kron', 'zKronFlux':'zflux_kron', 'yKronFlux':np.nan,
        'gKronFluxErr':'ge_flux_kron', 'rKronFluxErr':'re_flux_kron', 'iKronFluxErr':'ie_flux_kron',
        'zKronFluxErr':'ze_flux_kron', 'yKronFluxErr':np.nan,
        'gPSFFlux':'gflux_psf',
        'rPSFFlux':'rflux_psf', 'iKronFlux':'iflux_psf', 'zKronFlux':'zflux_psf', 'yKronFlux':np.nan,
        'gPSFFluxErr':'ge_flux_psf', 'rKronFluxErr':'re_flux_psf', 'iKronFluxErr':'ie_flux_psf',
        'zPSFFluxErr':'ze_flux_psf', 'yKronFluxErr':np.nan, 'gpsfChiSq':'gchi2_psf',
        'rpsfChiSq':'rchi2_psf', 'ipsfChiSq':'ichi2_psf', 'zpsfChiSq':'zchi2_psf',
        'ypsfChiSq':np.nan, 'nDetections':'ngood', 'SkyMapper_StarClass':'rclass_star',
        'distance':'r_cntr', 'objName':'object_id',
        'g_elong':'gelong', 'g_a':'ga', 'g_b':'gb', 'g_pa':'gpa',
        'r_elong':'relong', 'r_a':'ra', 'r_b':'rb', 'r_pa':'rpa',
        'i_elong':'ielong', 'i_a':'ia', 'i_b':'ib', 'i_pa':'ipa',
        'z_elong':'zelong', 'z_a':'za', 'z_b':'zb', 'z_pa':'zpa'} #'class_star' should be added

    keepCols = []
    for band in 'griz':
        for rad in ['radius_petro', 'radius_frac20', 'radius_frac50', 'radius_frac90']:
            flag_mapping[band + rad] = band + rad
            keepCols.append(band + rad)

    df_cols = np.array(list(flag_mapping.values()))
    mapped_cols = np.array(list(flag_mapping.keys()))

    PS1_cols = np.array(['objName', 'objAltName1', 'objAltName2', 'objAltName3', 'objID',
           'uniquePspsOBid', 'ippObjID', 'surveyID', 'htmID', 'zoneID',
           'tessID', 'projectionID', 'skyCellID', 'randomID', 'batchID',
           'dvoRegionID', 'processingVersion', 'objInfoFlag', 'qualityFlag',
           'raStack', 'decStack', 'raStackErr', 'decStackErr', 'raMean',
           'decMean', 'raMeanErr', 'decMeanErr', 'epochMean', 'posMeanChisq',
           'cx', 'cy', 'cz', 'lambda', 'beta', 'l', 'b', 'nStackObjectRows',
           'nStackDetections', 'nDetections', 'ng', 'nr', 'ni', 'nz', 'ny',
           'uniquePspsSTid', 'primaryDetection', 'bestDetection',
           'gippDetectID', 'gstackDetectID', 'gstackImageID', 'gra', 'gdec',
           'graErr', 'gdecErr', 'gEpoch', 'gPSFMag', 'gPSFMagErr', 'gApMag',
           'gApMagErr', 'gKronMag', 'gKronMagErr', 'ginfoFlag', 'ginfoFlag2',
           'ginfoFlag3', 'gnFrames', 'gxPos', 'gyPos', 'gxPosErr', 'gyPosErr',
           'gpsfMajorFWHM', 'gpsfMinorFWHM', 'gpsfTheta', 'gpsfCore',
           'gpsfLikelihood', 'gpsfQf', 'gpsfQfPerfect', 'gpsfChiSq',
           'gmomentXX', 'gmomentXY', 'gmomentYY', 'gmomentR1', 'gmomentRH',
           'gPSFFlux', 'gPSFFluxErr', 'gApFlux', 'gApFluxErr', 'gApFillFac',
           'gApRadius', 'gKronFlux', 'gKronFluxErr', 'gKronRad', 'gexpTime',
           'gExtNSigma', 'gsky', 'gskyErr', 'gzp', 'gPlateScale',
           'rippDetectID', 'rstackDetectID', 'rstackImageID', 'rra', 'rdec',
           'rraErr', 'rdecErr', 'rEpoch', 'rPSFMag', 'rPSFMagErr', 'rApMag',
           'rApMagErr', 'rKronMag', 'rKronMagErr', 'rinfoFlag', 'rinfoFlag2',
           'rinfoFlag3', 'rnFrames', 'rxPos', 'ryPos', 'rxPosErr', 'ryPosErr',
           'rpsfMajorFWHM', 'rpsfMinorFWHM', 'rpsfTheta', 'rpsfCore',
           'rpsfLikelihood', 'rpsfQf', 'rpsfQfPerfect', 'rpsfChiSq',
           'rmomentXX', 'rmomentXY', 'rmomentYY', 'rmomentR1', 'rmomentRH',
           'rPSFFlux', 'rPSFFluxErr', 'rApFlux', 'rApFluxErr', 'rApFillFac',
           'rApRadius', 'rKronFlux', 'rKronFluxErr', 'rKronRad', 'rexpTime',
           'rExtNSigma', 'rsky', 'rskyErr', 'rzp', 'rPlateScale',
           'iippDetectID', 'istackDetectID', 'istackImageID', 'ira', 'idec',
           'iraErr', 'idecErr', 'iEpoch', 'iPSFMag', 'iPSFMagErr', 'iApMag',
           'iApMagErr', 'iKronMag', 'iKronMagErr', 'iinfoFlag', 'iinfoFlag2',
           'iinfoFlag3', 'inFrames', 'ixPos', 'iyPos', 'ixPosErr', 'iyPosErr',
           'ipsfMajorFWHM', 'ipsfMinorFWHM', 'ipsfTheta', 'ipsfCore',
           'ipsfLikelihood', 'ipsfQf', 'ipsfQfPerfect', 'ipsfChiSq',
           'imomentXX', 'imomentXY', 'imomentYY', 'imomentR1', 'imomentRH',
           'iPSFFlux', 'iPSFFluxErr', 'iApFlux', 'iApFluxErr', 'iApFillFac',
           'iApRadius', 'iKronFlux', 'iKronFluxErr', 'iKronRad', 'iexpTime',
           'iExtNSigma', 'isky', 'iskyErr', 'izp', 'iPlateScale',
           'zippDetectID', 'zstackDetectID', 'zstackImageID', 'zra', 'zdec',
           'zraErr', 'zdecErr', 'zEpoch', 'zPSFMag', 'zPSFMagErr', 'zApMag',
           'zApMagErr', 'zKronMag', 'zKronMagErr', 'zinfoFlag', 'zinfoFlag2',
           'zinfoFlag3', 'znFrames', 'zxPos', 'zyPos', 'zxPosErr', 'zyPosErr',
           'zpsfMajorFWHM', 'zpsfMinorFWHM', 'zpsfTheta', 'zpsfCore',
           'zpsfLikelihood', 'zpsfQf', 'zpsfQfPerfect', 'zpsfChiSq',
           'zmomentXX', 'zmomentXY', 'zmomentYY', 'zmomentR1', 'zmomentRH',
           'zPSFFlux', 'zPSFFluxErr', 'zApFlux', 'zApFluxErr', 'zApFillFac',
           'zApRadius', 'zKronFlux', 'zKronFluxErr', 'zKronRad', 'zexpTime',
           'zExtNSigma', 'zsky', 'zskyErr', 'zzp', 'zPlateScale',
           'yippDetectID', 'ystackDetectID', 'ystackImageID', 'yra', 'ydec',
           'yraErr', 'ydecErr', 'yEpoch', 'yPSFMag', 'yPSFMagErr', 'yApMag',
           'yApMagErr', 'yKronMag', 'yKronMagErr', 'yinfoFlag', 'yinfoFlag2',
           'yinfoFlag3', 'ynFrames', 'yxPos', 'yyPos', 'yxPosErr', 'yyPosErr',
           'ypsfMajorFWHM', 'ypsfMinorFWHM', 'ypsfTheta', 'ypsfCore',
           'ypsfLikelihood', 'ypsfQf', 'ypsfQfPerfect', 'ypsfChiSq',
           'ymomentXX', 'ymomentXY', 'ymomentYY', 'ymomentR1', 'ymomentRH',
           'yPSFFlux', 'yPSFFluxErr', 'yApFlux', 'yApFluxErr', 'yApFillFac',
           'yApRadius', 'yKronFlux', 'yKronFluxErr', 'yKronRad', 'yexpTime',
           'yExtNSigma', 'ysky', 'yskyErr', 'yzp', 'yPlateScale', 'distance', 'SkyMapper_StarClass',
           'g_elong','g_a','g_b','g_pa',
           'r_elong','r_a','r_b','r_pa',
           'i_elong','i_a','i_b','i_pa',
           'z_elong','z_a','z_b','z_pa'])

    for i in np.arange(len(df_cols)):
        if df_cols[i] == 'nan':
            fullDF[mapped_cols[i]] = np.nan
        else:
            fullDF[mapped_cols[i]] = fullDF[df_cols[i]]

    # save the plate scale of skymapper in each band
    for band in 'grizy':
        fullDF['%sPlateScale'%band] = 0.5
    fullDF['primaryDetection'] = 1
    fullDF['bestDetection'] = 1

    #dummy variable set so that no sources get cut by qualityFlag in PS1 (which isn't in SkyMapper)
    fullDF['qualityFlag'] = 0

    #dummy variable
    fullDF['ny'] = 1
    colSet = np.concatenate([list(flag_mapping.keys()), ['gPlateScale', 'rPlateScale',
        'iPlateScale', 'zPlateScale', 'yPlateScale', 'primaryDetection', 'bestDetection', 'qualityFlag', 'ny']])

    fullDF = fullDF[colSet]

    leftover = set(PS1_cols) - set(fullDF.columns.values)
    for col in leftover:
        fullDF[col] = np.nan

    # arrange in the correct order for combining with northern-hemisphere PS1 sources
    fullDF = fullDF[PS1_cols]
    fullDF.drop_duplicates(subset=['objID'], inplace=True)
    return fullDF

def getDR2_petrosianSizes(ra_arr, dec_arr, rad):
    """Retrieves petrosian radius information from DR2 for panstarrs sources.

    :param ra_arr: List of right ascension values, in degrees.
    :type ra_arr: list
    :param dec_arr: List of declination values, in degrees.
    :type dec_arr: list
    :param rad: The search radius, in arcseconds.
    :type rad: float
    :return: Dataframe containing dr2 petrosian radiuses.
    :rtype: Pandas DataFrame
    """

    if len(ra_arr) < 1:
        return

    petroList = []
    for i in np.arange(len(ra_arr)):
        query = """select st.objID, st.primaryDetection, st.gpetR90, st.rpetR90, st.ipetR90, st.zpetR90, st.ypetR90
        from fGetNearbyObjEq(%.3f,%.3f,%.1f/60.0) nb
        inner join StackPetrosian st on st.objID=nb.objid where st.primaryDetection = 1""" %(ra_arr[i], dec_arr[i], rad)

        jobs = mastcasjobs.MastCasJobs(context="PanSTARRS_DR2")
        tab = jobs.quick(query, task_name="PetrosianRadiusSearch")
        df_petro = tab.to_pandas()
        petroList.append(df_petro)

    df_petro_full = pd.concat(petroList)
    return df_petro_full

def getDR2_halfLightSizes(ra_arr, dec_arr, rad):
    """Retrieves half-light radius information from DR2 for panstarrs sources.

    :param ra_arr: List of right ascension values, in degrees.
    :type ra_arr: list
    :param dec_arr: List of declination values, in degrees.
    :type dec_arr: list
    :param rad: The search radius, in arcseconds.
    :type rad: float
    :return: Dataframe containing dr2 half-light radiuses.
    :rtype: Pandas DataFrame
    """

    if len(ra_arr) < 1:
        return

    halfLightList = []
    for i in np.arange(len(ra_arr)):
        query = """select st.objID, st.primaryDetection, st.gHalfLightRad, st.rHalfLightRad, st.iHalfLightRad,
        st.zHalfLightRad, st.yHalfLightRad from fGetNearbyObjEq(%.3f,%.3f,%.1f/60.0) nb
        inner join StackModelFitExtra st on st.objID=nb.objid where st.primaryDetection = 1""" %(ra_arr[i], dec_arr[i], rad)

        jobs = mastcasjobs.MastCasJobs(context="PanSTARRS_DR2")
        tab = jobs.quick(query, task_name="halfLightSearch")
        df_halfLight = tab.to_pandas()
        halfLightList.append(df_halfLight)

    df_halfLight_full = pd.concat(halfLightList)
    return df_halfLight_full



import numpy as np
from astropy.table import Table
from PIL import Image
from io import BytesIO
import pathlib
import os
import sys
import pandas as pd
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
import re
import astro_ghost
from astro_ghost.PS1QueryFunctions import *
from astro_ghost.hostMatching import build_ML_df
from astro_ghost.NEDQueryFunctions import getNEDInfo
from astro_ghost.SimbadQueryFunctions import getSimbadInfo
from astro_ghost.gradientAscent import gradientAscent
from astro_ghost.starSeparation import separateStars_STRM, separateStars_South
from astro_ghost.sourceCleaning import clean_dict, removePS1Duplicates, getColors, makeCuts
from astro_ghost.stellarLocus import calc_7DCD
from astro_ghost.DLR import chooseByDLR
import requests
import pickle
import glob
from datetime import datetime
from joblib import dump, load

def getGHOST(real=False, verbose=False, installpath='', clobber=False):
    """Downloads the GHOST database.

    :param real: If True, download the GHOST database. If False, write an empty files with
        relevant columns (so that every transient is manually associated).
    :type real: bool, optional
    :param verbose: If True, print debugging info.
    :type verbose: bool, optional
    :param installpath: Filepath where GHOST database will be installed.
    :type installpath: str
    :param clobber: If True, write new GHOST database even if it already exists at installpath.
    :type clobber: bool, optional
    """

    if not installpath:
        try:
            installpath = os.environ['GHOST_PATH']
        except:
            print("Couldn't find where you want to save GHOST. Setting location to package path...")
            installpath = astro_ghost.__file__
            installpath = installpath.split("/")[:-1]
            installpath = "/".join(installpath)
    if not os.path.exists(installpath + '/database'):
        os.mkdir(installpath + '/database')
    else:
        if os.path.exists(installpath + '/database/GHOST.csv') & (clobber == False):
            print("GHOST database already exists in the install path!")
            return
    if real:
        url = 'https://www.dropbox.com/s/a0fufc3827pfril/GHOST.csv?dl=1'
        r = requests.get(url)
        fname = installpath + '/database/GHOST.csv'
        open(fname , 'wb').write(r.content)
        if verbose:
            print("Successfully downloaded GHOST database.\n")
    else:
        #create dummy database
        colnames = ['objName', 'objAltName1', 'objAltName2', 'objAltName3', 'objID',
            'uniquePspsOBid', 'ippObjID', 'surveyID', 'htmID', 'zoneID',
            'tessID', 'projectionID', 'skyCellID', 'randomID', 'batchID',
            'dvoRegionID', 'processingVersion', 'objInfoFlag', 'qualityFlag',
            'raStack', 'decStack', 'raStackErr', 'decStackErr', 'raMean',
            'decMean', 'raMeanErr', 'decMeanErr', 'epochMean', 'posMeanChisq',
            'cx', 'cy', 'cz', 'lambda', 'beta', 'l', 'b', 'nStackObjectRows',
            'nStackDetections', 'nDetections', 'ng', 'nr', 'ni', 'nz', 'ny',
            'uniquePspsSTid', 'primaryDetection', 'bestDetection',
            'gippDetectID', 'gstackDetectID', 'gstackImageID', 'gra', 'gdec',
            'graErr', 'gdecErr', 'gEpoch', 'gPSFMag', 'gPSFMagErr', 'gApMag',
            'gApMagErr', 'gKronMag', 'gKronMagErr', 'ginfoFlag', 'ginfoFlag2',
            'ginfoFlag3', 'gnFrames', 'gxPos', 'gyPos', 'gxPosErr', 'gyPosErr',
            'gpsfMajorFWHM', 'gpsfMinorFWHM', 'gpsfTheta', 'gpsfCore',
            'gpsfLikelihood', 'gpsfQf', 'gpsfQfPerfect', 'gpsfChiSq',
            'gmomentXX', 'gmomentXY', 'gmomentYY', 'gmomentR1', 'gmomentRH',
            'gPSFFlux', 'gPSFFluxErr', 'gApFlux', 'gApFluxErr', 'gApFillFac',
            'gApRadius', 'gKronFlux', 'gKronFluxErr', 'gKronRad', 'gexpTime',
            'gExtNSigma', 'gsky', 'gskyErr', 'gzp', 'gPlateScale',
            'rippDetectID', 'rstackDetectID', 'rstackImageID', 'rra', 'rdec',
            'rraErr', 'rdecErr', 'rEpoch', 'rPSFMag', 'rPSFMagErr', 'rApMag',
            'rApMagErr', 'rKronMag', 'rKronMagErr', 'rinfoFlag', 'rinfoFlag2',
            'rinfoFlag3', 'rnFrames', 'rxPos', 'ryPos', 'rxPosErr', 'ryPosErr',
            'rpsfMajorFWHM', 'rpsfMinorFWHM', 'rpsfTheta', 'rpsfCore',
            'rpsfLikelihood', 'rpsfQf', 'rpsfQfPerfect', 'rpsfChiSq',
            'rmomentXX', 'rmomentXY', 'rmomentYY', 'rmomentR1', 'rmomentRH',
            'rPSFFlux', 'rPSFFluxErr', 'rApFlux', 'rApFluxErr', 'rApFillFac',
            'rApRadius', 'rKronFlux', 'rKronFluxErr', 'rKronRad', 'rexpTime',
            'rExtNSigma', 'rsky', 'rskyErr', 'rzp', 'rPlateScale',
            'iippDetectID', 'istackDetectID', 'istackImageID', 'ira', 'idec',
            'iraErr', 'idecErr', 'iEpoch', 'iPSFMag', 'iPSFMagErr', 'iApMag',
            'iApMagErr', 'iKronMag', 'iKronMagErr', 'iinfoFlag', 'iinfoFlag2',
            'iinfoFlag3', 'inFrames', 'ixPos', 'iyPos', 'ixPosErr', 'iyPosErr',
            'ipsfMajorFWHM', 'ipsfMinorFWHM', 'ipsfTheta', 'ipsfCore',
            'ipsfLikelihood', 'ipsfQf', 'ipsfQfPerfect', 'ipsfChiSq',
            'imomentXX', 'imomentXY', 'imomentYY', 'imomentR1', 'imomentRH',
            'iPSFFlux', 'iPSFFluxErr', 'iApFlux', 'iApFluxErr', 'iApFillFac',
            'iApRadius', 'iKronFlux', 'iKronFluxErr', 'iKronRad', 'iexpTime',
            'iExtNSigma', 'isky', 'iskyErr', 'izp', 'iPlateScale',
            'zippDetectID', 'zstackDetectID', 'zstackImageID', 'zra', 'zdec',
            'zraErr', 'zdecErr', 'zEpoch', 'zPSFMag', 'zPSFMagErr', 'zApMag',
            'zApMagErr', 'zKronMag', 'zKronMagErr', 'zinfoFlag', 'zinfoFlag2',
            'zinfoFlag3', 'znFrames', 'zxPos', 'zyPos', 'zxPosErr', 'zyPosErr',
            'zpsfMajorFWHM', 'zpsfMinorFWHM', 'zpsfTheta', 'zpsfCore',
            'zpsfLikelihood', 'zpsfQf', 'zpsfQfPerfect', 'zpsfChiSq',
            'zmomentXX', 'zmomentXY', 'zmomentYY', 'zmomentR1', 'zmomentRH',
            'zPSFFlux', 'zPSFFluxErr', 'zApFlux', 'zApFluxErr', 'zApFillFac',
            'zApRadius', 'zKronFlux', 'zKronFluxErr', 'zKronRad', 'zexpTime',
            'zExtNSigma', 'zsky', 'zskyErr', 'zzp', 'zPlateScale',
            'yippDetectID', 'ystackDetectID', 'ystackImageID', 'yra', 'ydec',
            'yraErr', 'ydecErr', 'yEpoch', 'yPSFMag', 'yPSFMagErr', 'yApMag',
            'yApMagErr', 'yKronMag', 'yKronMagErr', 'yinfoFlag', 'yinfoFlag2',
            'yinfoFlag3', 'ynFrames', 'yxPos', 'yyPos', 'yxPosErr', 'yyPosErr',
            'ypsfMajorFWHM', 'ypsfMinorFWHM', 'ypsfTheta', 'ypsfCore',
            'ypsfLikelihood', 'ypsfQf', 'ypsfQfPerfect', 'ypsfChiSq',
            'ymomentXX', 'ymomentXY', 'ymomentYY', 'ymomentR1', 'ymomentRH',
            'yPSFFlux', 'yPSFFluxErr', 'yApFlux', 'yApFluxErr', 'yApFillFac',
            'yApRadius', 'yKronFlux', 'yKronFluxErr', 'yKronRad', 'yexpTime',
            'yExtNSigma', 'ysky', 'yskyErr', 'yzp', 'yPlateScale', 'distance',
            'NED_name', 'NED_type', 'NED_vel', 'NED_redshift', 'NED_mag',
            'i-z', 'g-r', 'r-i', 'g-i', 'z-y', 'g-rErr', 'r-iErr', 'i-zErr',
            'z-yErr', 'gApMag_gKronMag', 'rApMag_rKronMag', 'iApMag_iKronMag',
            'zApMag_zKronMag', 'yApMag_yKronMag', '7DCD', 'class', 'dist/DLR',
            'dist', 'TransientClass', 'TransientRA', 'TransientDEC',
            'TransientDiscoveryDate', 'TransientDiscoveryMag',
            'TransientRedshift', 'TransientDiscoveryYear', 'Transient AltName',
            'host_logmass', 'host_logmass_min', 'host_logmass_max',
            'Hubble Residual', 'TransientName']
        df = pd.DataFrame(columns = colnames)
        df.to_csv(installpath + "/database/GHOST.csv",index=False)
        if verbose:
            print("Successfully created dummy database.\n")

def fracWithHosts(transient_dict):
    """Calculates the fraction of supernovae
       in our dictionary that have at least
       one candidate host associated with
       them.

    :param transient_dict: The dictionary of supernovae and their host galaxy candidate objIDs from PS1.
    :type transient_dict: dictionary
    :return: The fraction of supernovae with at least one candidate host galaxy.
    :rtype: dictionary
    """

    count = 0
    for name, host in transient_dict.items():
        # only do matching if there's a found host
        if isinstance(host, list) or isinstance(host, np.ndarray):
            if len(host) > 0 and np.any(np.array(host == host)):
                count += 1
        else:
            if host == host:
                count+=1
    return count/len(transient_dict.keys())

def remove_prefix(text, prefix):
    """Removes the prefix from a string. Very useful for removing the 'SN' from
       supernova names!

    :param text: The full text.
    :type text: str
    :param prefix: The prefix to remove from the text.
    :type prefix: str
    :return: The input text, with prefix removed.
    :rtype: str
    """

    return text[text.startswith(prefix) and len(prefix):]

def getDBHostFromTransientCoords(transientCoords, GHOSTpath=''):
    """Gets the host associated with a supernova in
       the GHOST database by the supernova's position, if it exists.

    :param transientCoords: A list of astropy SkyCoord coordinates of transients.
    :type transientCoords: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str

    :return: The PS1 objects associated with the queried transients in GHOST.
    :rtype: Pandas DataFrame
    :return: A list of the coordinates of transients not found in the database.
    :rtype: host_DF : array-like
    """

    fullTable = fullData(GHOSTpath)
    notFound = []
    host_DF = None
    hostList = []

    #a little wrapper so that this function can take lists of coords in addition to one coord
    for transientCoord in transientCoords:
        smallTable = fullTable[np.abs(fullTable['TransientRA'] - transientCoord.ra.degree)<0.1]
        smallTable = smallTable[np.abs(smallTable['TransientDEC'] - transientCoord.dec.degree)<0.1]
        if len(smallTable) < 1:
            notFound.append(transientCoord)
            continue
        c2 = SkyCoord(smallTable['TransientRA']*u.deg, smallTable['TransientDEC']*u.deg, frame='icrs')
        sep = np.array(transientCoord.separation(c2).arcsec)
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host = smallTable.iloc[[host_idx]]
            hostList.append(host)
        else:
            notFound.append(transientCoord)
    if len(hostList) > 0:
        host_DF = pd.concat(hostList, ignore_index=True)
    return host_DF, notFound

def getDBHostFromTransientName(transientNames, GHOSTpath=''):
    """Gets the host associated with a supernova in
       the GHOST database by the supernova's name, if it exists.

    :param transientNames: A list of transient names.
    :type transientNames: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :return: The PS1 objects associated with the queried transients in GHOST.
    :rtype: Pandas DataFrame
    :return: A list of the coordinates of transients not found in the database.
    :rtype: array-like
    """

    fullTable = fullData(GHOSTpath)
    allHosts = []
    notFound = []
    host_DF = None
    for transientName in transientNames:
        transientName = re.sub(r"\s+", "", str(transientName))
        host = None
        possibleNames = [transientName, transientName.upper(), transientName.lower(), "SN"+transientName]
        for name in possibleNames:
            if len(fullTable[fullTable['TransientName'] == name])>0:
                host = fullTable[fullTable['TransientName'] == name]
                allHosts.append(host)
                break
        if host is None:
            notFound.append(transientName)
    if len(allHosts) > 0:
        host_DF = pd.concat(allHosts, ignore_index=True)
    return host_DF, notFound

def getHostFromHostName(hostNames, GHOSTpath=''):
    """Gets hosts in the GHOST database by name.

    :param hostNames: A list of host galaxy names.
    :type hostNames: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :return: The host galaxies found in GHOST.
    :rtype: Pandas DataFrame
    """

    fullTable = fullData(GHOSTpath)
    possibleNames = []

    for name in hostNames:
        possibleNames.append(name)
        possibleNames.append(name.upper())
        possibleNames.append(re.sub(r"\s+", "", str(name)))
        possibleNames.append(name.lower())

    host = fullTable[fullTable['NED_name'].isin(possibleNames)]
    if host is None:
        print("Sorry, no hosts were found in our database!\n")
    return host

def getHostFromHostCoords(hostCoords, GHOSTpath=''):
    """Gets hosts in the GHOST database by coordinates.

    :param hostCoords : A list of astropy SkyCoord coordinates of host galaxies.
    :type hostCoords: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :return: The subset of discovered hosts.
    :rtype: Pandas DataFrame
    """

    fullTable = fullData(GHOSTpath)
    c2 = SkyCoord(fullTable['raMean']*u.deg, fullTable['decMean']*u.deg, frame='icrs')
    host = []
    for hostCoord in hostCoords:
        sep = np.array(hostCoord.separation(c2).arcsec)
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host.append(fullTable.iloc[[host_idx]])

    if len(host) < 1:
        print("Sorry, No hosts found in our database!\n")
    else:
        host = pd.concat(host, ignore_index=True)
    return host

def getTransientStatsFromHostCoords(hostCoord):
    """Prints basic statistics for transient,
       based on a query of the coordinates of
       its host.

    :param hostCoord: The position of the host galaxy.
    :type hostCoord: Astropy SkyCoord Object
    """

    host = getHostFromHostCoords(hostCoord)
    i = 0
    if len(host) > 0:
        print("Found info for host %s.\n"%host['NED_name'].values[0])
        print("Associated supernovae: ")
        for SN in np.array(host['TransientName'].values):
            SN_frame = host.loc[host['TransientName'] == SN]
            print("%i. %s"%(i+1, SN_frame['TransientName'].values[0]))
            print("RA, DEC (J2000): %f, %f"%(SN_frame['TransientRA'].values[0], SN_frame['TransientDEC'].values[0]))
            print("Redshift: %f"%SN_frame['TransientRedshift'].values[0])
            print("Discovery Date: %s"%SN_frame['TransientDiscoveryDate'].values[0].split(" ")[0])
            print("Discovery Mag: %.2f"%SN_frame['TransientDiscoveryMag'].values[0])
            i+= 1
    return

def getTransientStatsFromHostName(hostName):
    """Returns basic statistics for transient,
       based on a query of the name of
       its host.

    :param hostName: The name of the host galaxy.
    :type hostName: str
    """

    host = getHostFromHostName(hostName)
    hostCoord = SkyCoord(np.unique(host['raMean'])*u.deg, np.unique(host['decMean'])*u.deg, frame='icrs')
    getTransientStatsFromHostCoords(hostCoord)
    return

def getHostStatsFromTransientCoords(transientCoordsList, GHOSTpath=''):
    """ Returns basic statistics for the most likely
        host of a previously identified transient, from the
        transient's coordinates.

    :param transientCoordsList: A list of astropy SkyCoord coordinates of transients.
    :type transientCoordsList: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    """

    fullTable = fullData(GHOSTpath)
    names = []
    for transientCoords in transientCoordsList:
        c2 = SkyCoord(fullTable['TransientRA']*u.deg, fullTable['TransientDEC']*u.deg, frame='icrs')
        sep = np.array(transientCoords.separation(c2).arcsec)
        host = None
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host = fullTable.iloc[[host_idx]]
            names.append(host['TransientName'].values[0])
    getHostStatsFromTransientName(names)

def getHostStatsFromTransientName(transientName, GHOSTpath=''):
    """Returns basic statistics for the most likely host of a previously identified transient.

    :param transientName: Array of transient names.
    :type transientName: array-like
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    """

    transientName = np.array(transientName)
    fullTable = fullData(GHOSTpath)
    host, notFound = getDBHostFromTransientName(transientName, GHOSTpath)
    if host is not None:
        for idx, row in host.iterrows():
            if np.unique(row['NED_name']) != "":
                print("Found host %s.\n"%row['NED_name'])
            else:
                print("Found host with PS1 ID %s"%row['objID'])
            print("RA, DEC (J2000): %f, %f"%(row['raMean'], row['decMean']))
            if row['NED_redshift'] != '':
                print("Redshift: %f"%row['NED_redshift'])
            print("PS1 rMag: %.2f"%row['rApMag'])
            print("g-r          r-i          i-z")
            print("%.2f+/-%.3f   %.2f+/-%.3f   %.2f+/-%.3f"%(row['g-r'],row['g-rErr'], row['r-i'], row['r-iErr'], row['i-z'], row['i-zErr']))
            print("Associated supernovae: ")
            if np.unique(row['NED_name']) != "":
                tempHost = fullTable[fullTable['NED_name'] == row['NED_name']]
            else:
                tempHost = fullTable[fullTable['objID'] == row['objID']]
            for SN in np.array(tempHost['TransientName'].values):
                print(SN)
    else:
        print("No host info found!")
    return

def getHostImage(transientName='', band="grizy", rad=60, save=False, GHOSTpath=''):
    """Returns a postage stamp of the most likely host in one of the PS1 bands - g,r,i,z,y - as a
       fits file with radius rad, and plots the image.

    :param transientName: Name of queried transient.
    :type transientName: str
    :param band: Band for host-galaxy image.
    :type band: str
    :param rad: Size of the image, in arcsec.
    :type rad: float
    :param save: If True, save host image.
    :type save: bool, optional
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    """

    if transientName == '':
        print("Error! Please enter a supernova!\n")
        return
    fullTable = fullData(GHOSTpath)
    host, notFound = getDBHostFromTransientName(transientName, GHOSTpath)
    if host is not None:
        host.reset_index(drop=True, inplace=True)
        tempSize = int(4*float(rad))
        fn_save = host['objID'].values[0]
        if np.unique(host['NED_name']) != "":
            fn_save = host['NED_name'].values[0]
            print("Showing postage stamp for %s"%np.unique(host['NED_name'])[0])
        ra = np.unique(host['raMean'])[0]
        dec = np.unique(host['decMean'])[0]
        tempSize = int(4*rad)
        img = getcolorim(ra, dec, output_size=tempSize, size=tempSize, filters=band, format="png")
        plt.figure(figsize=(10,10))
        plt.imshow(img)
        if save:
            img.save("%s.png" % fn_save)
        return
    else:
        print("Transient host not found!")

def getTransientSpectra(path, transientName):
    """Retrieves all saved spectra associated with a transient.

    :param path: Filepath to spectra.
    :type path: str
    :param transientName: Name of transient to query.
    :type transientName: str
    :return: Saved spectra.
    :rtype: list of Pandas DataFrames
    """

    transientName = remove_prefix(transientName, 'SN')
    files = glob.glob(path+"*%s*"%transientName)
    specFiles = []
    if len(files) > 0:
        for file in files:
            if remove_prefix(file, path).startswith("osc_"):
                spectra = pd.read_csv(file, header=None, names=['Wave(A)', 'I', 'Ierr'])
            else:
                spectra = pd.read_csv(file, delim_whitespace=True, header=None, names=['x', 'y', 'z'])
            specFiles.append(spectra)
    else:
        print("Sorry! No spectra found.")
    return specFiles

def getHostSpectra(transientName, path):
    """Retrieves all saved spectra associated with a host galaxy.

    :param path: Filepath to spectra.
    :type path: str
    :param transientName: Name of transient to query.
    :type transientName: str
    :return: Saved spectra.
    :rtype: list of Pandas DataFrames
    """

    transientName = remove_prefix(transientName, 'SN')
    files = glob.glob(path+"*%s_hostSpectra.csv*"%transientName)
    specFiles = []
    if len(files) > 0:
        for file in files:
            spectra = pd.read_csv(file)
            specFiles.append(spectra)
    else:
        print("Sorry! No spectra found.")
    return specFiles

def coneSearchPairs(coord, radius, GHOSTpath=''):
    """A cone search for all transient-host pairs within a certain radius, returned as a pandas dataframe.

    :param coord: Astropy SkyCoord
    :type coord: Position for cone search.
    :param radius: Search radius, in arcsec.
    :type radius: float
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :return: GHOST galaxies within search radius.
    :rtype: Pandas DataFrame
    """

    fullTable = fullData(GHOSTpath)
    c2 = SkyCoord(fullTable['TransientRA']*u.deg, fullTable['TransientDEC']*u.deg, frame='icrs')
    sep = np.array(coord.separation(c2).arcsec)
    hosts = None
    if np.nanmin(sep) < radius:
        host_idx = np.where(sep < radius)[0]
        hosts = fullTable.iloc[host_idx]
    return hosts

def fullData(GHOSTpath=''):
    """Returns the full GHOST database.

    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :return: GHOST database.
    :rtype: Pandas DataFrame
    """

    if not GHOSTpath:
        GHOSTpath = os.getenv('GHOST_PATH')
        if not GHOSTpath:
            try:
                GHOSTpath = astro_ghost.__file__
                GHOSTpath = GHOSTpath.split("/")[:-1]
                GHOSTpath = "/".join(GHOSTpath)
            except:
                print("Error! I don't know where you installed GHOST -- set GHOST_PATH as an environmental variable or pass in the GHOSTpath parameter.")
    fullTable = pd.read_csv(GHOSTpath+"/database/GHOST.csv")
    return fullTable

def getTransientHosts(transientName=[''], snCoord=[''], snClass=[''], verbose=False, starcut='normal', ascentMatch=False, px=800, savepath='./', GHOSTpath='', redo_search=True):
    """The wrapper function for the main host association pipeline. The function first
       searches the pre-existing GHOST database by transient name, then by transient coordinates, and finally
       associates the remaining transients not found.

    :param transientName: List of transients to associate.
    :type transientName: array-like
    :param snCoord: List of astropy SkyCoord transient positions.
    :type snCoord: array-like
    :param snClass: List of transient classifications (if they exist).
    :type snClass: array-like
    :param verbose: If True, print logging information.
    :type verbose: bool, optional
    :param starcut: Strings corresponding to the classification thresholds required to classify a star as such.
        Options are \\'gentle\\' (P>0.8), normal (P>0.5), and aggressive (P>0.3).
    :type starcut: str
    :param ascentMatch: If True, run the gradient ascent algorithm for the transients not matched with the
        Directional Light Radius algorithm.
    :type ascentMatch: bool
    :param px: Size of the image used in gradient ascent (ignored if ascentMatch=False).
    :type px: int
    :param savepath: Filepath where dataframe of associated hosts will be saved.
    :type savepath: str
    :param GHOSTpath: The path to the saved GHOST database.
    :type GHOSTpath: str
    :param redo_search: If True, redo the search with 150\\" cone search radius if hosts were not
        found for any of the queried transients.
    :type redo_search: bool
    :return: Final dataframe of associated transients and host galaxies.
    :rtype: Pandas DataFrame
    """

    #if no names were passed in, add placeholder names for each transient in the search
    if len(transientName) < 1:
        transientName = []
        print("No transient names listed, adding placeholder names...")
        for i in np.arange(len(snCoord)):
            transientName.append("Transient_%i" % (i+1))
    hostDB = None
    tempHost1 = tempHost2 = tempHost3 = None
    found_by_name = found_by_coord = found_by_manual = 0
    if not isinstance(snCoord, list) and not isinstance(snCoord, np.ndarray):
        snCoord = [snCoord]
        transientName = [transientName]
        snClass = [snClass]
    if len(snClass) != len(transientName):
        snClass = ['']*len(transientName)

    transientName = [x.replace(" ", "") for x in transientName]
    df_transients = pd.DataFrame({'Name':np.array(transientName), 'snCoord':np.array(snCoord), 'snClass':np.array(snClass)})

    tempHost1, notFoundNames = getDBHostFromTransientName(transientName, GHOSTpath)
    found_by_name = len(transientName) - len(notFoundNames)

    if tempHost1 is None or len(notFoundNames) > 0:
        if verbose:
            print("%i transients not found in GHOST by name, trying a coordinate search..."%len(notFoundNames))

        df_transients_remaining = df_transients[df_transients['Name'].isin(notFoundNames)]

        snCoord_remaining =  df_transients_remaining['snCoord'].values

        tempHost2, notFoundCoords = getDBHostFromTransientCoords(snCoord_remaining, GHOSTpath);
        found_by_coord = len(transientName) - len(notFoundCoords) - found_by_name
        if tempHost2 is None or len(notFoundCoords) > 0:
            if verbose:
                print("%i transients not found in GHOST by name or coordinates, manually associating..."% len(notFoundCoords))

            df_transients_remaining = df_transients_remaining[df_transients_remaining['snCoord'].isin(notFoundCoords)]

            transientName_remaining =  df_transients_remaining['Name'].values
            snCoord_remaining = df_transients_remaining['snCoord'].values
            snClass_remaining = df_transients_remaining['snClass'].values

            tempHost3 = findNewHosts(transientName_remaining, snCoord_remaining, snClass_remaining, verbose, starcut, ascentMatch, px, savepath)

            if (len(transientName_remaining) > 0) and (len(tempHost3)==0) and (redo_search):
                 #bump up the search radius to 150 arcsec for extremely low-redshift hosts...
                 if verbose:
                     print("Couldn't find any hosts! Trying again with a search radius of 150''.")
                 tempHost3 = findNewHosts(transientName_remaining, snCoord_remaining, snClass_remaining, verbose, starcut, ascentMatch, px, savepath, 150)
            found_by_manual = len(tempHost3)
    hostDB = pd.concat([tempHost1, tempHost2, tempHost3], ignore_index=True)
    hostDB.replace(-999.0, np.nan, inplace=True)
    if verbose:
        print("%i transients found by name, %i transients found by coordinates, %i transients manually associated."% (found_by_name, found_by_coord, found_by_manual))
    return hostDB

def findNewHosts(transientName, snCoord, snClass, verbose=False, starcut='gentle', ascentMatch=False, px=800, savepath='./', rad=60):
    """Associates hosts of transients not in the GHOST database.

    :param transientName: List of transients to associate.
    :type transientName: array-like
    :param snCoord: List of astropy SkyCoord transient positions.
    :type snCoord: array-like
    :param snClass: List of transient classifications (if they exist).
    :type snClass: array-like
    :param verbose: If True, print logging information.
    :type verbose: bool, optional
    :param starcut: Strings corresponding to the classification thresholds required to classify a star as such.
        Options are \\'gentle\\' (P>0.8), normal (P>0.5), and aggressive (P>0.3).
    :type starcut: str, optional
    :param ascentMatch: If True, run the gradient ascent algorithm for the transients not matched with the
        Directional Light Radius algorithm.
    :type ascentMatch: bool, optional
    :param px: Size of the image used in gradient ascent (ignored if ascentMatch=False).
    :type px: int
    :param savepath: Filepath where dataframe of associated hosts will be saved.
    :type savepath: str
    :param rad: The search radius around each transient position, in arcseconds.
    :type rad: float
    :return: Final dataframe of associated transients and host galaxies.
    :rtype: Pandas DataFrame
    """

    if isinstance(transientName, str):
        transientName = transientName.replace(" ", "")
        snRA = snCoord.ra.degree
        snDEC = snCoord.dec.degree
    elif isinstance(transientName, list) or isinstance(transientName, np.ndarray):
        transientName = [x.replace(" ", "") for x in transientName]
        snRA = [x.ra.degree for x in snCoord]
        snDEC = [x.dec.degree for x in snCoord]
    else:
        print("Error! Please pass in your transient name as either a string or a list/array of strings.\n")

    transientName_arr = np.array(transientName)
    snRA_arr = np.array(snRA)
    snDEC_arr = np.array(snDEC)
    snClass_arr = np.array(snClass)

    dateStr = str(datetime.today()).replace("-", '').replace(".", '').replace(":", "").replace(" ", '')

    fn_Host = "SNe_TNS_%s_PS1Hosts_%iarcsec.csv" % (dateStr, rad)
    fn_SN = 'transients_%s.csv' % dateStr
    fn_Dict = fn_Host[:-4] + ".p"
    dir_name = fn_SN[:-4]
    if not os.path.exists(savepath):
        os.mkdir(savepath)
        os.chmod(savepath, 0o777)
    os.makedirs(savepath + dir_name)
    path = savepath+dir_name+'/'

    #create temp dataframe with RA and DEC corresponding to the transient
    snDF = pd.DataFrame({'Name':transientName_arr, 'RA':snRA_arr, 'DEC':snDEC_arr, 'HostName':['']*len(snDEC_arr), 'Obj. Type':snClass_arr})
    snDF.to_csv(path+fn_SN, index=False)

    #begin doing the heavy lifting to associate transients with hosts
    host_DF = get_hosts(path, fn_SN, fn_Host, rad)
    #galaxySizes = getDR2_forcedGalaxySizes(snRA_arr, snDEC_arr, rad)
    #host_DF_new = host_DF.merge(galaxySizes, on='objID',how='outer')
    #host_DF = host_DF_new if not galaxySizes.empty else host_DF

    if len(host_DF) < 1:
        print("ERROR: Found no hosts in cone search during manual association!")
        return None

    cuts = ["n", "coords", "duplicate"]

    transient_dict =[]
    # this bit of trickery is required to combine northern-hemisphere and
    # southern-hemisphere source dictionaries
    f = open(path+'/dictionaries/'+fn_Dict, "rb")
    try:
        while True:
            transient_dict.append(pickle.load(f))
    except EOFError:
        pass
    temp = transient_dict[0]
    if len(transient_dict) > 1:
        for i in np.arange(len(transient_dict)-1):
            temp.update(transient_dict[i+1])
    transient_dict = {k.replace(' ', ''): v for k, v in temp.items()}
    desperate_dict = transient_dict.copy()

    host_DF = getNEDInfo(host_DF)

    host_DF_north = host_DF[host_DF['decMean']>-30].reset_index(drop=True)
    host_DF_south = host_DF[host_DF['decMean']<=-30].reset_index(drop=True)

    host_DF_north = makeCuts(host_DF_north, cuts, transient_dict)

    host_DF = pd.concat([host_DF_north, host_DF_south], ignore_index=True)

    cut_dict = clean_dict(transient_dict, host_DF, [])
    desperate_dict = cut_dict.copy()

    hostFrac = fracWithHosts(cut_dict)*100
    if verbose:
        print("Associated fraction after quality cuts: %.2f%%."%hostFrac)

    #automatically add to host association for gradient ascent
    lost = np.array([k for k, v in cut_dict.items() if len(v) <1])

    host_DF_north = getColors(host_DF_north)
    host_DF_north = removePS1Duplicates(host_DF_north)
    host_DF_north = calc_7DCD(host_DF_north)

    host_DF = pd.concat([host_DF_north, host_DF_south], ignore_index=True)
    host_DF = getSimbadInfo(host_DF)
    host_DF.to_csv(path+"/candidateHosts_NEDSimbad.csv", index=False)

    host_DF_north = host_DF[host_DF['decMean']>-30].reset_index()
    host_DF_south = host_DF[host_DF['decMean']<=-30].reset_index()

    host_gals_DF_north, stars_DF_north = separateStars_STRM(host_DF_north, plot=0, verbose=verbose, starcut=starcut)
    host_gals_DF_south, stars_DF_south = separateStars_South(host_DF_south, plot=0, verbose=verbose, starcut=starcut)
    host_gals_DF = pd.concat([host_gals_DF_north, host_gals_DF_south],ignore_index=True)
    stars_DF = pd.concat([stars_DF_north, stars_DF_south],ignore_index=True)

    if verbose:
        print("Removed %i stars. We now have %i candidate host galaxies."%(len(stars_DF), len(host_gals_DF)))

    cut_dict = clean_dict(cut_dict, host_gals_DF, [])
    stars_DF.to_csv(path+"removedStars.csv",index=False)

    #plotLocus(host_gals_DF, color=1, save=1, type="Gals", timestamp=dateStr)
    #plotLocus(stars_DF, color=1, save=1, type="Stars", timestamp=dateStr)

    host_dict_nospace = {k.replace(' ', ''): v for k, v in cut_dict.items()}

    fn = "DLR.txt"
    transients = pd.read_csv(path+fn_SN)

    with open(path+"/dictionaries/checkpoint_preDLR.p", 'wb') as fp:
           dump(host_dict_nospace, fp)

    #return the candidates list before DLR for debugging purposes
    #return host_gals_DF

    host_DF, host_dict_nospace_postDLR, noHosts, GD_SN = chooseByDLR(path, host_gals_DF, transients, fn, host_dict_nospace, todo="r")

    #last-ditch effort -- for the ones with no found host, just pick the nearest NED galaxy.
    for transient in noHosts:
          tempDF = host_gals_DF[host_gals_DF['objID'].isin(desperate_dict[transient])]
          tempDF_gals = tempDF[tempDF['NED_type'].isin(['G', 'PofG', 'GPair', 'GGroup', 'GClstr'])].reset_index()
          if len(tempDF_gals) < 1:
             continue
          transientRA = transients.loc[transients['Name'] == transient, 'RA'].values[0]
          transientDEC = transients.loc[transients['Name'] == transient, 'DEC'].values[0]
          transientCoord = SkyCoord(transientRA*u.deg, transientDEC*u.deg, frame='icrs')
          tempHostCoords = SkyCoord(tempDF_gals['raMean'].values*u.deg, tempDF_gals['decMean'].values*u.deg, frame='icrs')
          sep = transientCoord.separation(tempHostCoords)
          desperateMatch = tempDF_gals.iloc[[np.argmin(sep.arcsec)]]
          host_DF = pd.concat([host_DF, desperateMatch], ignore_index=True)
          host_dict_nospace_postDLR[transient] = desperateMatch['objID'].values[0]
          if verbose:
               print("Desperate match found for %s, %.2f arcsec away." % (transient, sep[np.argmin(sep.arcsec)].arcsec))

    if len(noHosts) > 0:
        with open(path+"/dictionaries/noHosts_fromDLR.p", 'wb') as fp:
            dump(noHosts, fp)

    if len(GD_SN) > 0:
        with open(path+"/dictionaries/badInfo_fromDLR.p", 'wb') as fp:
             dump(GD_SN, fp)

    #gradient ascent algorithm for the SNe that didn't pass this stage
    SN_toReassociate = np.concatenate([np.array(noHosts), np.array(GD_SN), np.array(list(lost))])

    if (len(SN_toReassociate) > 0) and (ascentMatch):
        if verbose:
            print("%i transients with no host found with DLR, %i transients with bad host data with DLR." %(len(noHosts), len(GD_SN)))
            print("Running gradient ascent for %i remaining transients."%len(SN_toReassociate))
            print("See GradientAscent.txt for more information.")

        fn_GD= path+'/GradientAscent.txt'

        host_dict_nospace_postDLR_GD, host_DF, unchanged = gradientAscent(path, transient_dict,  host_dict_nospace_postDLR, SN_toReassociate, host_DF, transients, fn_GD, plot=verbose, px=px)

        with open(path+"/dictionaries/gals_postGD.p", 'wb') as fp:
            dump(host_dict_nospace_postDLR_GD, fp)

        if verbose:
            print("Hosts not found for %i transients in gradient ascent. Storing names in GD_unchanged.txt" %(len(unchanged)))

        with open(path+"/GD_unchanged.txt", 'wb') as fp:
            dump(unchanged, fp)

        hostFrac = fracWithHosts(host_dict_nospace_postDLR_GD)*100

        if verbose:
            print("Associated fraction after gradient ascent: %.2f%%."%hostFrac)

        final_dict = host_dict_nospace_postDLR_GD.copy()

    else:
        final_dict = host_dict_nospace_postDLR.copy()

    host_DF = build_ML_df(final_dict, host_DF, transients)

    with open(path+"/dictionaries/" + "Final_Dictionary.p", 'wb') as fp:
           dump(final_dict, fp)

    #a few final cleaning steps
    host_DF.drop_duplicates(subset=['TransientName'], inplace=True)
    host_DF = host_DF[host_DF['TransientName'] != ""]
    host_DF.reset_index(inplace=True, drop=True)
    host_DF['TransientName'] = [x.replace(" ", "") for x in host_DF['TransientName']]

    matchFrac = len(host_DF)/len(transients)*100
    print("Found matches for %.1f%% of events."%matchFrac)
    if verbose:
        print("Saving table of hosts to %s."%(path+"tables/FinalAssociationTable.csv"))

    host_DF.to_csv(path+"FinalAssociationTable.csv", index=False)

    tablePath = path+"/tables/"
    printoutPath = path+'/printouts/'

    paths = [tablePath, printoutPath]

    for tempPath in paths:
        if not os.path.exists(tempPath):
            os.mkdir(tempPath)

    #move tables to the tables directory, and printouts to the printouts directory
    printouts = glob.glob(path+'/*.txt')
    for p in printouts:
        fn = remove_prefix(p, path)
        os.rename(p, printoutPath+fn)

    table_ext = ['csv', 'gz']
    for end in table_ext:
        tables = glob.glob(path+'/*.%s'%end)
        for t in tables:
            fn = remove_prefix(t, path)
            os.rename(t, tablePath+fn)
    #remove if there's an extra index column
    try:
        del host_DF['index']
    except:
        pass
    return host_DF
